# Copyright 2024 StreamSets Inc.

# fmt: off
from copy import deepcopy

import pytest

from streamsets.sdk.sch import ControlHub
from streamsets.sdk.sch_models import JobSequence, JobSequenceBuilder

from .resources.conftest_data import JOB_SEQUENCE_BUILDER_JSON, JOB_SEQUENCE_EMPTY_JSON

# fmt: on

START_TIME_2099 = 4200268800
END_TIME_2099 = 4200268800
START_TIME_2100 = 4237401600
END_TIME_2100 = 4237996800
UTC_TIME_Z0NE = 'UTC'
BASIC_CRON_TAB_MASK = '0/1 * 1/1 * ? *'
NUM_OF_STEPS = 4
NUM_OF_JOBS = 3


class MockControlHub:
    def __init__(self):
        self.organization = 'DUMMY ORG'

    @property
    def _sequencing_api(self):
        return {'definitions': {'USequence': deepcopy(JOB_SEQUENCE_BUILDER_JSON)}}


def test_get_job_sequence_builder_build():
    job_sequence_builder = ControlHub.get_job_sequence_builder(MockControlHub())
    assert isinstance(job_sequence_builder, JobSequenceBuilder)
    assert isinstance(job_sequence_builder._control_hub, MockControlHub)

    assert job_sequence_builder._job_sequence == deepcopy(JOB_SEQUENCE_EMPTY_JSON)

    name, description = 'TEST NAME', 'TEST DESC'
    job_sequence = job_sequence_builder.build(name=name, description=description)
    assert isinstance(job_sequence, JobSequence)
    assert isinstance(job_sequence._control_hub, MockControlHub)
    assert job_sequence.name == name
    assert job_sequence.description == description


def test_get_job_sequence_builder_build_with_incorrect_types():
    job_sequence_builder = ControlHub.get_job_sequence_builder(MockControlHub())

    name, description = 'TEST NAME', 'TEST DESC'

    with pytest.raises(TypeError):
        job_sequence_builder.build(name=123, description=description)

    with pytest.raises(TypeError):
        job_sequence_builder.build(name=name, description=123)


def test_get_job_sequence_builder_build_with_default_values():
    job_sequence_builder = ControlHub.get_job_sequence_builder(MockControlHub())
    job_sequence = job_sequence_builder.build()
    assert job_sequence.name == 'Job Sequence'
    assert job_sequence.description is None


def test_get_job_sequence_builder_add_start_condition():
    job_sequence_builder = ControlHub.get_job_sequence_builder(MockControlHub())
    assert job_sequence_builder._job_sequence == deepcopy(JOB_SEQUENCE_EMPTY_JSON)

    job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)
    assert job_sequence_builder._job_sequence['startTime'] == START_TIME_2099
    assert job_sequence_builder._job_sequence['endTime'] == END_TIME_2099
    assert job_sequence_builder._job_sequence['timezone'] == UTC_TIME_Z0NE
    assert job_sequence_builder._job_sequence['crontabMask'] == BASIC_CRON_TAB_MASK


def test_get_job_sequence_builder_add_start_condition_incorrect_types():
    job_sequence_builder = ControlHub.get_job_sequence_builder(MockControlHub())

    with pytest.raises(TypeError):
        job_sequence_builder.add_start_condition('STRING', END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)

    with pytest.raises(TypeError):
        job_sequence_builder.add_start_condition(START_TIME_2099, 'STRING', UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)

    with pytest.raises(TypeError):
        job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, 123, BASIC_CRON_TAB_MASK)

    with pytest.raises(TypeError):
        job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, 123)


def test_get_job_sequence_builder_add_start_condition_with_default_values():
    job_sequence_builder = ControlHub.get_job_sequence_builder(MockControlHub())

    job_sequence_builder.add_start_condition()
    assert job_sequence_builder._job_sequence['startTime'] == 0
    assert job_sequence_builder._job_sequence['endTime'] == 0
    assert job_sequence_builder._job_sequence['timezone'] == UTC_TIME_Z0NE
    assert job_sequence_builder._job_sequence['crontabMask'] == BASIC_CRON_TAB_MASK



# Copyright 2019 StreamSets Inc.

"""Models to be used by multiple StreamSets components."""

# fmt: off
import json
import warnings
from uuid import uuid4

import inflection

# fmt: on

json_to_python_style = lambda x: inflection.underscore(x)
python_to_json_style = lambda x: inflection.camelize(x, uppercase_first_letter=False)


class Configuration:
    """Abstraction for configurations.

    This class enables easy access to and modification of data stored as a list of dictionaries. A Configuration is
    stored in the form:

    .. code-block:: none

        [{"name" : "<name_1>","value" : "<value_1>"}, {"name" : "<name_2>", value" : "<value_2>"},...]

    However, the passed in configuration parameter can be a list of Configurations such as:

    .. code-block:: none

        [[{"name" : "<name_1>","value" : "<value_1>"}, {"name" : "<name_2>", value" : "<value_2>"},...],
        [{"name" : "<name_3>","value" : "<value_3>"}, {"name" : "<name_4>", value" : "<value_4>"},...],...]

    Args:
        compatibility_map (:obj:`dict`, optional): A dictionary mapping values used for backwards compatibility.
        configuration (:obj:`list`): List of configurations (see above for format).
        property_key (:obj:`str`, optional): The dictionary entry denoting the property key.
            Default: ``name``
        property_value (:obj:`str`, optional): The dictionary entry denoting the property value.
            Default: ``value``
        update_callable (optional): A callable to which ``self._data`` will be passed as part of ``__setitem__``.
        update_callable_kwargs (:obj:`dict`, optional): A dictionary of kwargs to pass (along with a body)
            to the callable.
        id_to_remap (:obj:`dict`, optional): A dictionary mapping configuration IDs to human-readable container keys.
                                             Example: {'custom_region':'googleCloudConfig.customRegion', ... }
    """

    # Use an uber secret class attribute to specify whether other attributes can be assigned by __setattr__.
    __frozen = False

    def __init__(
        self,
        configuration=None,
        compatibility_map=None,
        property_key='name',
        property_value='value',
        update_callable=None,
        update_callable_kwargs=None,
        id_to_remap=None,
    ):
        # Apply overrides to initial data
        if compatibility_map:
            for configuration_entry in configuration:
                configuration_name = configuration_entry.get(property_key)
                configuration_value = configuration_entry.get(property_value)

                if configuration_name in compatibility_map:
                    overrides = compatibility_map[configuration_name]
                    override_values = overrides.get('values', {})

                    configuration_entry[property_key] = overrides['name']

                    if configuration_value in override_values:
                        configuration_entry[property_value] = override_values[configuration_value]
                        warnings.warn(
                            'Configuration {}={} has been deprecated. Please use {}={} instead.'.format(
                                configuration_name,
                                configuration_value,
                                overrides['name'],
                                override_values[configuration_value],
                            ),
                            DeprecationWarning,
                        )
                    else:
                        warnings.warn(
                            'Configuration {} has been deprecated. Please use {} instead.'.format(
                                configuration_name, overrides['name']
                            ),
                            DeprecationWarning,
                        )

        self._compatibility_map = compatibility_map or {}
        self.property_key = property_key
        self.property_value = property_value

        self._id_to_remap = id_to_remap or {}
        self._update_callable = update_callable
        self._update_callable_kwargs = update_callable_kwargs or {}

        # Ensure the input 'configuration' is properly formatted, handling both single configurations and lists.
        self._data = [configuration] if isinstance(configuration[0], dict) else configuration
        self._configuration_index_map = self._create_configuration_index_map()

        self.__frozen = True

    def _create_configuration_index_map(self):
        """
        Creates a mapping {config_item_name: (config_item_index, config_list_index)} for efficient lookups.
        This method facilitates quick identification of the index of a configuration item within the
        Configurations/self._data list.

        Here,
        1. config_list_index is the index within the Configurations list, i.e. the outer index.
        2. config_item_index is the index within an individual Configuration, i.e. the inner index.

        If our ._data is as follows:
            [[{"name" : "<name_1>","value" : "<value_1>"}, {"name" : "<name_2>", value" : "<value_2>"}],
            [{"name" : "<name_3>","value" : "<value_3>"}, {"name" : "<name_4>", value" : "<value_4>"}]]

        Then configuration_index_map would be returned as:
            {<name_1>:(0,0), <name_2>:(1,0),, <name_3>:(0,1),, <name_4>:(1,1),}

        Returns:
            A configuration index map (:obj:`dict`).

        Raises:
            TypeError: If the input data structure is not a list of configurations,
                       if a Configuration is not a list of dictionaries, or
                       if a Configuration does not contain the specified property_key and property_value.
        """
        configuration_index_map = {}
        for config_list_index, config_list in enumerate(self._data):
            if not isinstance(config_list, list):
                raise TypeError('Please pass in a list of configurations')

            for config_item_index, config_item in enumerate(config_list):
                if not isinstance(config_item, dict):
                    raise TypeError('A Configuration must be a list of dictionaries')
                if self.property_key not in config_item or self.property_value not in config_item:
                    raise TypeError(
                        'Configuration {} does not contain property_key:{} or property_value:{}'.format(
                            config_item, self.property_key, self.property_value
                        )
                    )
                configuration_index_map[config_item[self.property_key]] = (config_item_index, config_list_index)

        return configuration_index_map

    def __getattr__(self, key):
        if not self.__frozen:
            super().__getattr__(key)
            return

        return self.__getitem__(key)

    def __getitem__(self, key):
        if key in self._id_to_remap:
            key = self._id_to_remap[key]

        if key not in self._configuration_index_map:
            raise AttributeError(key)

        index, configuration_index = self._configuration_index_map[key]
        config = self._data[configuration_index][index]
        return self._convert_value(config)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        if not self.__frozen:
            super().__setattr__(key, value)
            return

        if key in self._id_to_remap:
            key = self._id_to_remap[key]

        if key in self._compatibility_map:
            overrides = self._compatibility_map[key]
            if 'values' in overrides and value in self._compatibility_map[key]['values']:
                warnings.warn(
                    'Deprecation warning: Configuration {}={} is deprecated on this engine version. '
                    'Updating value to {}={}.'.format(key, value, overrides['name'], overrides['values'][value]),
                    DeprecationWarning,
                )
                value = overrides['values'][value]
            else:
                warnings.warn(
                    'Configuration {} has been deprecated. Please use {} instead.'.format(key, overrides['name']),
                    DeprecationWarning,
                )

            key = overrides['name']

        if key not in self._configuration_index_map:
            raise AttributeError(key)

        index, configuration_index = self._configuration_index_map[key]
        config = self._data[configuration_index][index]

        config[self.property_value] = value
        if self._update_callable:
            kwargs = dict(body=[config])
            kwargs.update(self._update_callable_kwargs)
            self._update_callable(**kwargs)

    def __contains__(self, item):
        return item in self._id_to_remap or item in self._configuration_index_map

    def __repr__(self):
        configs = {}
        for configuration in self._data:
            for config in configuration:
                key = config[self.property_key]
                configs[key] = self._convert_value(config)

        # If a key has a remapped key, delete the original key and add the remapped key into configs
        for remapped_key, original_key in self._id_to_remap.items():
            if original_key != remapped_key and original_key in configs:
                configs[remapped_key] = configs[original_key]
                del configs[original_key]

        return '{{{}}}'.format(', '.join("'{}': {}".format(k, v) for k, v in configs.items()))

    def __dir__(self):
        # Stripping out any values that have multiple words in it. Example: 'Organization account type'
        # in sch_models.Organization.configuration
        id_to_remap_cleaned = [key for key in self._id_to_remap.keys() if ' ' not in key]
        return sorted(list(dir(object)) + list(self.__dict__.keys()) + id_to_remap_cleaned)

    def items(self):
        """Gets the configuration's items.

        Returns:
            A new view of the configuration’s items ((key, value) pairs).
        """
        # To keep the behavior in line with a Python dict's, we'll generate one and then use its items method.
        configuration_dict = {}
        for configuration in self._data:
            for config in configuration:
                configuration_dict[config[self.property_key]] = self._convert_value(config)

        for config_property in self._id_to_remap:
            key = self._id_to_remap[config_property]
            if key in configuration_dict:
                configuration_dict[config_property] = configuration_dict[key]
                del configuration_dict[key]
        return configuration_dict.items()

    def get(self, key, default=None):
        """Return the value of key or, if not in the configuration, the default value."""
        try:
            return self[key]
        except KeyError:
            return default

    def update(self, configs):
        """Update instance with a collection of configurations.

        Args:
            configs (:obj:`dict`): Dictionary of configurations to use.
        """
        for key, value in configs.items():
            self[key] = value

    def _convert_value(self, config):
        if config.get('type') == 'boolean':
            return json.loads(config[self.property_value])
        elif config.get('type') == 'integer':
            return int(config[self.property_value])
        else:
            return config[self.property_value]


class BaseModel:
    """Base class for StreamSets Accounts models that essentially just wrap a dictionary.

    Args:
        data (:obj:`dict`): The underlying JSON representation of the model.
        attributes_to_ignore (:obj:`list`, optional): A list of string attributes to mask from being handled
            by this class' __setattr__ method. Default: ``None``.
        attributes_to_remap (:obj:`dict`, optional): A dictionary of attributes to remap with the desired attributes
            as keys and the corresponding property name in the JSON representation as values. Default: ``None``.
        repr_metadata (:obj:`list`, optional): A list of attributes to use in the model's __repr__ string.
            Default: ``None``.
    """

    def __new__(cls, *args, **kwargs):
        instance = super(BaseModel, cls).__new__(cls)
        super(BaseModel, instance).__setattr__('_data_internal', {})
        super(BaseModel, instance).__setattr__('_attributes_to_ignore', [])
        super(BaseModel, instance).__setattr__('_attributes_to_remap', {})
        super(BaseModel, instance).__setattr__('_repr_metadata', [])

        return instance

    def __init__(self, data, attributes_to_ignore=None, attributes_to_remap=None, repr_metadata=None):
        # _data_internal is introduced to  help inherited classes that need to load _data when _data is accessed
        # eg. Pipeline
        super().__setattr__('_data_internal', data)
        super().__setattr__('_attributes_to_ignore', attributes_to_ignore or [])
        super().__setattr__('_attributes_to_remap', attributes_to_remap or {})
        super().__setattr__('_repr_metadata', repr_metadata or [])

    # By default these properties don't do anything by can be overrided by inherited classes to load something
    @property
    def _data_internal(self):
        return self.__dict__['_data'] if '_data' in self.__dict__ else None

    @_data_internal.setter
    def _data_internal(self, data):
        self.__dict__['_data'] = data

    @property
    def _data(self):
        return self._data_internal

    @_data.setter
    def _data(self, data):
        self._data_internal = data

    def __getattr__(self, name):
        name_ = python_to_json_style(name)
        if name in self._attributes_to_remap:
            remapped_name = self._attributes_to_remap[name]
            return self._data_internal[remapped_name]
        elif (
            name_ in self._data_internal
            and name not in self._attributes_to_ignore
            and name not in self._attributes_to_remap.values()
        ):
            return self._data_internal[name_]
        raise AttributeError('Could not find attribute {}.'.format(name_))

    def __setattr__(self, name, value):
        name_ = python_to_json_style(name)
        if name in self._attributes_to_remap:
            remapped_name = self._attributes_to_remap[name]
            self._data_internal[remapped_name] = value
        elif (
            name_ in self._data_internal
            and name not in self._attributes_to_ignore
            and name not in self._attributes_to_remap.values()
        ):
            self._data_internal[name_] = value
        else:
            super().__setattr__(name, value)

    def __dir__(self):
        return sorted(
            list(dir(object))
            + list(self.__dict__.keys())
            + list(
                json_to_python_style(key)
                for key in self._data_internal.keys()
                if key not in (list(self._attributes_to_remap.values()) + self._attributes_to_ignore)
            )
            + list(self._attributes_to_remap.keys())
        )

    def __eq__(self, other):
        return self._data_internal == other._data_internal

    def __repr__(self):
        return '<{} ({})>'.format(
            self.__class__.__name__, ', '.join('{}={}'.format(key, getattr(self, key)) for key in self._repr_metadata)
        )


class ModelCollection:
    """Base class wrapper with Abstractions.

    Args:
        streamsets_entity: An instance of underlysing StreamSets entity
            e.g. :py:class:`streamsets.sdk.next.Next` or :py:class:`streamsets.sdk.accounts.Accounts`.
    """

    def __init__(self, streamsets_entity):
        self._streamsets_entity = streamsets_entity
        self._id_attr = 'id'

    def _get_all_results_from_api(self, **kwargs):
        """Used to get multiple (all) results from api.

        Args:
            Optional arguments to be passed to filter the results.

        Returns:
            A :obj:`collections.namedtuple`: of
                results (:py:class:`streamsets.sdk.utils.SeekableList`): a SeekableList of inherited instances of
                :py:class:`streamsets.sdk.models.BaseModel` and
                kwargs (:obj:`dict`): a dict of local variables not used in this function.
        """
        pass

    def __iter__(self):
        """Enables the list enumeration or iteration."""
        for item in self._get_all_results_from_api().results:
            yield item

    def __getitem__(self, i):
        """Enables the user to fetch items by index.

        Args:
            i (:obj:`int`): Index of the item.

        Returns:
            An inherited instance of :py:class:`streamsets.sdk.models.BaseModel`.
        """
        return self._get_all_results_from_api().results[i]

    def __len__(self):
        """Provides length (count) of items.

        Returns:
            A :py:obj:`int` object.
        """
        return len(self._get_all_results_from_api().results)

    def __contains__(self, item_given):
        """Checks if given item is in the list of items by comparing the ids.

        Returns:
            A :py:obj:`boolean` object.
        """
        return self.contains(**{self._id_attr: getattr(item_given, self._id_attr)})

    def get(self, **kwargs):
        """
        Args:
            **kwargs: Optional arguments to be passed to filter the results offline.

        Returns:
            An inherited instance of :py:class:`streamsets.sdk.models.BaseModel`.
        """
        result, new_kwargs = self._get_all_results_from_api(**kwargs)
        return result.get(**new_kwargs)

    def get_all(self, **kwargs):
        """
        Args:
            **kwargs: Optional other arguments to be passed to filter the results offline.

        Returns:
            A :py:obj:`streamsets.sdk.utils.SeekableList` of inherited instances of
            :py:class:`streamsets.sdk.models.BaseModel`.
        """
        result, new_kwargs = self._get_all_results_from_api(**kwargs)
        return result.get_all(**new_kwargs)

    def __repr__(self):
        return str(self._get_all_results_from_api().results)

    def contains(self, **kwargs):
        """
        Args:
            **kwargs: Optional arguments to be passed to filter the results offline.

        Returns:
            A :py:obj:`boolean` object.
        """
        try:
            self.get(**kwargs)
        except ValueError:
            return False
        return True


class _StageWithPredicates:
    """Pipeline Stage extension to include a predicate configuration and handle the output lanes.
    This class is expected to inherit from either SDC or ST Stage class.

    Args:
            variable_output_drive (:obj:`str`, optional): Configuration name that drives the output lanes available.
            Default: ``None``.
    """

    def __init__(self, variable_output_drive=None, *args, **kwargs):
        self.variable_output_drive = variable_output_drive
        super().__init__(*args, **kwargs)
        if not self.predicates:
            self.predicates = ['default']

    def _prepare_predicates(self, predicates):
        """Validate a configuration dictionary and add an output lane if necessary.

        Args:
            predicates (:obj:`list`): A list of predicates in predicate/outputLane form or as predicates.

        Returns:
            A corresponding :obj:`list` of dictionaries containing the predicate and outputLane for each entry.
        """

        formatted_predicates = []
        for predicate in predicates:
            if not isinstance(predicate, dict):
                predicate = {'predicate': str(predicate)}

            output_lane = predicate.get(
                'outputLane', '{}OutputLane{}'.format(self._data['instanceName'], str(uuid4())).replace('-', '_')
            )

            if not predicate.get('predicate', None):
                raise ValueError('Output Lane drives should have a predicate key.')
            if not predicate.get('outputLane', None):
                output_lane_config = {'outputLane': output_lane}

                # We make sure that the predicate is in the form {'outputLane': 'value', 'predicate': 'value'} to be
                # consistent with the UI
                predicate = {**output_lane_config, **predicate}
            formatted_predicates.append(predicate)
        return formatted_predicates

    def _set_config(self, config_property, value):
        # Check whether the configuration asked for is driving the output lanes.
        if self.variable_output_drive == config_property.config_name:
            if isinstance(value, list):
                self.predicates = value
            else:
                raise ValueError('Configuration must be a list of predicates or a list of dictionaries.')
        else:
            super()._set_config(config_property, value)

    def _disconnect_and_remove_predicate(self, i):
        deleted_predicate = self.predicates.pop(i)
        output_lane = deleted_predicate['outputLane']

        output_lane_index = self.output_lanes.index(output_lane)
        self.disconnect_output_lanes(output_lane_index=output_lane_index)

        self._data['outputLanes'].remove(output_lane)
        self.output_streams -= 1

    @property
    def predicates(self):
        """
        Get the predicate list for this stage.
        Predicates define the output lanes number and behaviour.
        Set it using either a list of predicates or a list of the full dictionary specifying the output lane.

        Example:
            stage.predicates = ['>0']
            stage.predicates = ['>0', 'default']
            stage.predicates = [{'predicate':'>0', 'outputLane': 'lane1'}]

        Returns:
            A :obj:`list` of the predicates.
        """
        return self.configuration[self.variable_output_drive]

    @predicates.setter
    def predicates(self, predicates):
        predicates = self._prepare_predicates(predicates)

        # Create default condition if not present
        if not [predicate for predicate in predicates if predicate['predicate'] == 'default']:
            default_predicate = self._prepare_predicates(['default'])
            predicates.extend(default_predicate)

        if not predicates[-1]['predicate'] == 'default':
            raise ValueError('The default predicate must be placed at the end of the list.')

        self.disconnect_output_lanes(all_stages=True)

        self.output_streams = 0
        self._data['outputLanes'] = []
        self._output_lane_idx = 0

        # Create output lanes in parent stage for each predicate
        for predicate in predicates:
            self._data['outputLanes'].append(predicate['outputLane'])
            self.output_streams += 1

        self.configuration[self.variable_output_drive] = predicates

    def add_predicates(self, predicates):
        """Add a predicate.

        Example:
            stage.add_predicates(['>0'])
            stage.add_predicates([{'predicate':'>0', 'outputLane':'lane1'}])
            stage.add_predicates(['>0' ,'=0'])

        Args:
            predicates (:obj:`list`): The list of predicates to add.

        """
        if not isinstance(predicates, list):
            raise ValueError('Predicates should be a list.')

        formatted_predicates = self._prepare_predicates(predicates)

        for new_predicate in formatted_predicates:
            self._data['outputLanes'].insert(0, new_predicate['outputLane'])
            self.predicates.insert(0, new_predicate)
            self.output_streams += 1

    def remove_predicate(self, predicate):
        """Remove a predicate.

        Example:
            stage.remove_predicate(stage.predicates[0])
            stage.remove_predicate({'predicate':'>0', 'outputLane':'lane1'})

        Args:
            predicate (:obj:`dict`): The predicate to delete as a dictionary including the outputLane.

        """

        clean_predicate = next(iter(self._prepare_predicates([predicate])), {})

        if not clean_predicate:
            raise ValueError("Need to specify a predicate")

        if clean_predicate.get('predicate') == 'default':
            raise ValueError("Can't delete the default predicate.")

        for i, found_predicate in enumerate(self.predicates):
            if found_predicate == clean_predicate:
                self._disconnect_and_remove_predicate(i)
                return
        raise ValueError("Can't find target predicate in the predicates list.")



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2024

from copy import deepcopy

import pytest

from streamsets.sdk import ControlHub
from streamsets.sdk.constants import SDC_EXECUTOR_TYPE
from streamsets.sdk.sch_models import Connection

from .resources.connections_data import CONNECTION_INTERNAL_JSON


class DummyEngine:
    def __init__(self, _id):
        self.id = _id
        self.engine_type = SDC_EXECUTOR_TYPE


class DummyEngines:
    def get(self, id):
        return DummyEngine(id)


class DummyControlHub(ControlHub):
    def __init__(self, *args, **kwargs):
        # purposely avoid calling super().__init__()
        # we want access to the method and not the initializing steps
        pass

    @property
    def engines(self):
        return DummyEngines()


@pytest.fixture(scope="function")
def dummy_sch():
    yield DummyControlHub()


@pytest.fixture(scope="function")
def mock_connection(dummy_sch):
    connection_internal_data = deepcopy(CONNECTION_INTERNAL_JSON)
    connection = Connection(connection=connection_internal_data, control_hub=dummy_sch)
    yield connection


@pytest.mark.parametrize("not_a_connection", [None, 'banana'])
def test_verify_connection_raises_type_error_for_invalid_connection(dummy_sch, not_a_connection):
    with pytest.raises(TypeError):
        dummy_sch.verify_connection(not_a_connection)


def test_verify_connection_raises_type_error_for_invalid_library_type(dummy_sch, mock_connection):
    with pytest.raises(TypeError):
        dummy_sch.verify_connection(connection=mock_connection, library={"this": "isn't", "a": "string"})


def test_verify_connection_raises_value_error_for_invalid_library(dummy_sch, mock_connection):
    with pytest.raises(ValueError):
        dummy_sch.verify_connection(connection=mock_connection, library='not jdbc')



# Copyright 2023 StreamSets Inc.

# fmt: off
import json

import pytest

from streamsets.sdk.sch_models import JobStatus
from streamsets.sdk.utils import get_random_string

# fmt: on

NUM_PIPELINES = 2
NUM_JOBS = 2
START_TIME_2099 = 4200268800
END_TIME_2099 = 4200854400
UTC_TIME_Z0NE = 'UTC'
BASIC_CRON_TAB_MASK = '0/1 * 1/1 * ? *'

# TODO: Add tests for more saql_search_types than just PIPELINE


@pytest.fixture(scope="module")
def sample_saql_pipeline_searches(sch):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    pipeline_searches = []
    for i in range(NUM_PIPELINES):
        pipeline_searches.append(
            sch.api_client.create_saql_pipeline_search(
                {
                    "name": "SAMPLE_PIPELINE_SEARCH_{}_{}".format(i, get_random_string()),
                    "type": "PIPELINE",
                    "mode": "BASIC",
                    "query": "name==*tmp_query*",
                }
            ).response.json()
        )
    try:
        yield pipeline_searches
    finally:
        for search in pipeline_searches:
            sch.api_client.remove_saql_pipeline_search(search["id"])


@pytest.fixture(scope="module")
def sample_saql_job_searches(sch):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    job_searches = []
    for i in range(NUM_PIPELINES):
        job_searches.append(
            sch.api_client.create_saql_job_search(
                {
                    "name": "SAMPLE_JOB_SEARCH_{}_{}".format(i, get_random_string()),
                    "type": "JOB_INSTANCE",
                    "mode": "BASIC",
                    "query": "name==*tmp_query*",
                }
            ).response.json()
        )
    try:
        yield job_searches
    finally:
        for search in job_searches:
            sch.api_client.remove_saql_job_search(search["id"])


@pytest.fixture(scope="module")
def sample_saql_search_object(sch):
    """
    Trivial SAQL Search builder:
    """
    # build the SAQLSearch Object
    builder = sch.get_saql_search_builder('PIPELINE')
    builder.add_filter(
        property_name='name',
        property_operator='contains',
        property_value='TEST_NAME',
        property_condition_combiner='AND',
    )
    obj = builder.build('TEST NAME {}'.format(get_random_string()))

    yield obj


@pytest.fixture(scope="module")
def simple_pipeline(sch, sch_authoring_sdc_id):
    """A trivial pipeline:

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.api_client.delete_pipeline(pipeline.pipeline_id)


@pytest.fixture(scope="module")
def sample_jobs(sch, simple_pipeline, sch_executor_sdc_label):
    """A set of simple jobs based on simple pipeline."""
    job_builder = sch.get_job_builder()

    jobs = []
    for i in range(5):
        job = job_builder.build('test_simple_job_fetch_{}'.format(i), pipeline=simple_pipeline)
        job.data_collector_labels = sch_executor_sdc_label
        sch.add_job(job)
        jobs.append(job)

    try:
        yield jobs
    finally:
        for job in jobs:
            sch.delete_job(job)


def test_saql_pipeline_search_api(sch, sample_saql_pipeline_searches):
    pipeline_searches = sch.api_client.get_saql_pipeline_searches(
        "SAMPLE_PIPELINE_SEARCH_1", "PIPELINE"
    ).response.json()
    assert pipeline_searches["totalCount"] == 1


def test_saql_job_search_api(sch, sample_saql_job_searches):
    job_searches = sch.api_client.get_saql_job_searches("SAMPLE_JOB_SEARCH_1", "JOB_INSTANCE").response.json()
    assert job_searches["totalCount"] == 1


def test_saql_favorite_job_api(sch, sample_saql_job_searches):
    job_searches = sch.api_client.get_saql_job_searches("SAMPLE_JOB_SEARCH_1", "JOB_INSTANCE").response.json()
    assert job_searches["totalCount"] == 1

    search = job_searches['data'][0]
    sch.api_client.create_saql_fav_job(search['id']).response.json()

    favorite_job = sch.api_client.get_saql_job_searches("SAMPLE_JOB_SEARCH_1", "JOB_INSTANCE").response.json()
    assert favorite_job["totalCount"] == 1
    assert favorite_job['data'][0]['favorite']['favorite']


def test_saql_favorite_pipeline_api(sch, sample_saql_pipeline_searches):
    job_searches = sch.api_client.get_saql_pipeline_searches("SAMPLE_PIPELINE_SEARCH_1", "PIPELINE").response.json()
    assert job_searches["totalCount"] == 1
    search = job_searches['data'][0]

    sch.api_client.create_saql_fav_pipeline(search['id']).response.json()

    favorite_pipeline = sch.api_client.get_saql_pipeline_searches(
        "SAMPLE_PIPELINE_SEARCH_1", "PIPELINE"
    ).response.json()
    assert favorite_pipeline["totalCount"] == 1
    assert favorite_pipeline['data'][0]['favorite']['favorite']


def test_saql_saved_searches_pipeline(sch):
    try:
        response = sch.api_client.create_saql_pipeline_search(
            {"name": "TEST_PIPELINE_SEARCH", "type": "PIPELINE", "mode": "BASIC", "query": "name==*tmp_query*"}
        ).response.json()
        saql_searches = sch.saql_saved_searches_pipeline.get_all(name='TEST_PIPELINE_SEARCH')
        assert len(saql_searches) == 1
    finally:
        sch.api_client.remove_saql_pipeline_search(response["id"])


def test_saql_saved_searches_fragment(sch):
    try:
        response = sch.api_client.create_saql_pipeline_search(
            {"name": "SAMPLE_FRAGMENT_SEARCH", "type": "FRAGMENT", "mode": "BASIC", "query": "name==*tmp_query*"}
        ).response.json()
        saql_searches = sch.saql_saved_searches_fragment.get_all(name='SAMPLE_FRAGMENT_SEARCH')
        assert len(saql_searches) == 1
    finally:
        sch.api_client.remove_saql_pipeline_search(response["id"])


def test_saql_saved_searches_job_instance(sch):
    try:
        response = sch.api_client.create_saql_job_search(
            {
                "name": "SAMPLE_JOB_INSTANCE_SEARCH",
                "type": "JOB_INSTANCE",
                "mode": "BASIC",
                "query": "name==*tmp_query*",
            }
        ).response.json()
        saql_searches = sch.saql_saved_searches_job_instance.get_all(name='SAMPLE_JOB_INSTANCE_SEARCH')
        assert len(saql_searches) == 1
    finally:
        sch.api_client.remove_saql_job_search(response["id"])


def test_saql_saved_searches_job_template(sch):
    try:
        response = sch.api_client.create_saql_job_search(
            {
                "name": "SAMPLE_JOB_TEMPLATE_SEARCH",
                "type": "JOB_TEMPLATE",
                "mode": "BASIC",
                "query": "name==*tmp_query*",
            }
        ).response.json()
        saql_searches = sch.saql_saved_searches_job_template.get_all(name='SAMPLE_JOB_TEMPLATE_SEARCH')
        assert len(saql_searches) == 1
    finally:
        sch.api_client.remove_saql_job_search(response["id"])


def test_saql_saved_searches_job_draft_run(sch):
    try:
        response = sch.api_client.create_saql_job_search(
            {
                "name": "SAMPLE_JOB_DRAFT_RUN_SEARCH",
                "type": "JOB_DRAFT_RUN",
                "mode": "BASIC",
                "query": "name==*tmp_query*",
            }
        ).response.json()
        saql_searches = sch.saql_saved_searches_draft_run.get_all(name='SAMPLE_JOB_DRAFT_RUN_SEARCH')
        assert len(saql_searches) == 1
    finally:
        sch.api_client.remove_saql_job_search(response["id"])


def test_saql_saved_searches_job_run(sch, sample_jobs):
    job = sample_jobs[0]
    sch.start_job(job)
    sch.stop_job(job)

    saql_search_builder = sch.get_saql_search_builder(saql_search_type='JOB_RUN', mode='ADVANCED')

    saql_search_builder.query = "color ==GRAY"
    saql_search_object = saql_search_builder.build(name='TEST JOB RUN QUERY')
    history_results = job.job_history.get_all(search=saql_search_object)

    for result in history_results:
        assert isinstance(result, JobStatus)
        assert result.color == 'GRAY'

    saql_search_builder.query = "color ==RED"
    saql_search_object = saql_search_builder.build(name='TEST JOB RUN QUERY')
    history_results = job.job_history.get_all(search=saql_search_object)

    for result in history_results:
        assert isinstance(result, JobStatus)
        assert result.color == 'RED'


def test_saql_saved_searches_job_sequence(sch):
    try:
        random_string = get_random_string()
        sequences = []
        for i in range(5):
            job_sequence_builder = sch.get_job_sequence_builder()

            job_sequence_builder.add_start_condition(START_TIME_2099, END_TIME_2099, UTC_TIME_Z0NE, BASIC_CRON_TAB_MASK)

            job_sequence = job_sequence_builder.build(
                name='{} {}'.format(random_string, i),
                description='description {}'.format(random_string),
            )
            sch.publish_job_sequence(job_sequence=job_sequence)

            sequences.append(job_sequence.id)

        saql_search_builder = sch.get_saql_search_builder(saql_search_type='JOB_SEQUENCE', mode='ADVANCED')

        saql_search_builder.query = 'name == "*{}*"'.format(random_string)
        saql_search_object = saql_search_builder.build(name='TEST JOB SEQUENCE QUERY')

        results = sch.job_sequences.get_all(search=saql_search_object)

        assert len(results) == 5
        for result in results:
            assert result.id in sequences

    except Exception as e:
        raise e
    finally:
        for sequence_id in sequences:
            sch.api_client.delete_job_sequence(sequence_id)


def test_saql_builder(sch):
    # create/test builder with filters
    builder = sch.get_saql_search_builder('PIPELINE')
    assert len(builder.query) == 0
    builder.add_filter(
        property_name='name',
        property_operator='contains',
        property_value='TEST_NAME',
        property_condition_combiner='AND',
    )
    assert len(builder.query) == 1

    # test builder object
    obj = builder.build('TEST NAME')
    assert obj.name == 'TEST NAME'
    assert obj.type == 'PIPELINE'
    assert isinstance(obj.query, str)


def test_saql_builder_with_incorrect_saql_search_type(sch):
    try:
        sch.get_saql_search_builder(saql_search_type='FOO', mode='BASIC')
    except Exception as exception:
        assert isinstance(exception, ValueError)


def test_saql_builder_with_incorrect_saql_search_mode(sch):
    try:
        sch.get_saql_search_builder(saql_search_type='PIPELINE', mode='foo')
    except Exception as exception:
        assert isinstance(exception, ValueError)


def test_saql_builder_with_incorrect_property_name(sch):
    builder = sch.get_saql_search_builder('PIPELINE')
    try:
        builder.add_filter(
            property_name='foo',
            property_operator='contains',
            property_value='TEST_NAME',
            property_condition_combiner='AND',
        )
    except Exception as exception:
        assert isinstance(exception, ValueError)


def test_saql_builder_with_incorrect_property_operator(sch):
    builder = sch.get_saql_search_builder('PIPELINE')
    try:
        builder.add_filter(
            property_name='name', property_operator='foo', property_value='TEST_NAME', property_condition_combiner='AND'
        )
    except Exception as exception:
        assert isinstance(exception, ValueError)


def test_saql_builder_with_incorrect_search_name_type(sch):
    builder = sch.get_saql_search_builder(saql_search_type='PIPELINE', mode='BASIC')
    builder.add_filter(
        property_name='name',
        property_operator='contains',
        property_value='TEST_NAME',
        property_condition_combiner='AND',
    )

    with pytest.raises(TypeError):
        builder.build(12)


def test_save_search(sch, sample_saql_search_object):
    response_save = {}
    try:
        # test basic save response details
        response = sch.save_saql_search(sample_saql_search_object)
        assert response.response.status_code == 200
        response_save = response.response.json()
        assert isinstance(response_save, dict)
        assert response_save['name'] == sample_saql_search_object.name
        assert response_save['type'] == 'PIPELINE'

        # test returned query
        response_json = json.loads(response_save['query'])
        assert len(response_json) == 1
        query = response_json[0]
        assert query['filter']['name'] == 'name'
        assert query['operator'] == 'contains'
        assert query['value'] == 'TEST_NAME'
        assert query['conditionCombiner'] == 'AND'
    finally:
        # remove the saql search
        if response_save.get("id", None):
            sch.api_client.remove_saql_pipeline_search(response_save["id"])


def test_mark_search_as_favorite(sch, sample_saql_search_object):
    response_save = {}
    try:
        response_save = sch.save_saql_search(sample_saql_search_object).response.json()

        # test marking a search as favorite
        saql_search = sch.saql_saved_searches_pipeline.get(name=sample_saql_search_object.name)
        response = sch.mark_saql_search_as_favorite(saql_search)
        assert response.response.status_code == 200
        response_favorite = response.response.json()
        assert response_favorite == saql_search.favorite
    finally:
        # remove the saql search
        if response_save.get("id", None):
            sch.api_client.remove_saql_pipeline_search(response_save["id"])


def test_remove_search(sch, sample_saql_search_object):
    response_save = {}
    try:
        response_save = sch.save_saql_search(sample_saql_search_object).response.json()

        # test remove a search
        saql_search = sch.saql_saved_searches_pipeline.get(name=sample_saql_search_object.name)
        response = sch.remove_saql_search(saql_search)
        assert response.response.status_code == 200
        response_remove = response.response.json()
        assert response_remove['id'] == response_save['id']
    except Exception as exception:
        # remove the saql search
        if response_save.get("id", None):
            sch.api_client.remove_saql_pipeline_search(response_save["id"])

        raise exception


def test_query_with_multiple_filters(sch):
    # build the SAQLSearch Object
    builder = sch.get_saql_search_builder('PIPELINE')

    builder.add_filter(
        property_name='name', property_operator='contains', property_value='SAMPLE', property_condition_combiner='AND'
    )
    builder.add_filter(
        property_name='name', property_operator='begins with', property_value='TEST', property_condition_combiner='AND'
    )
    builder.add_filter(
        property_name='modified_on',
        property_operator='is before',
        property_value='2023-04-03',
        property_condition_combiner='OR',
    )
    builder.add_filter(property_name='draft', property_operator='is true', property_condition_combiner='AND')
    builder.add_filter(property_name='commit_message', property_operator='is empty')

    saql_search_object = builder.build(name='TEST NAME')

    assert (
        saql_search_object.query == 'name == "*SAMPLE*" and name == "TEST*" and modified_on =lt= "2023-04-03"'
        ' or draft == true and commit_message =null= true'
    )

    response_save = {}
    try:
        # test basic save response details
        response = sch.save_saql_search(saql_search_object)
        assert response.response.status_code == 200
        response_save = response.response.json()
    finally:
        # remove the saql search
        if response_save.get("id", None):
            sch.api_client.remove_saql_pipeline_search(response_save["id"])



# Copyright 2023 StreamSets Inc.

# fmt: off
import json
from copy import copy, deepcopy

import pytest

from streamsets.sdk.sch_models import BaseModel

from .resources.base_model_data import BASE_MODEL_JSON

# fmt: on


@pytest.fixture(scope="function")
def base_model_data():
    data = deepcopy(BASE_MODEL_JSON)
    return data


@pytest.fixture(scope="module")
def attributes_to_ignore():
    return ['provenanceMetaData']


@pytest.fixture(scope="module")
def attributes_to_remap():
    # Mapping is {"new_attribute": "original_attribute"}
    return {'committed_by': 'committer', 'topology_name': 'name'}


@pytest.fixture(scope="module")
def repr_metadata():
    return ['topology_id', 'topology_name']


class Helper(BaseModel):
    def __init__(self, data, base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
        super().__init__(
            data=base_model_data,
            attributes_to_ignore=attributes_to_ignore,
            attributes_to_remap=attributes_to_remap,
            repr_metadata=repr_metadata,
        )
        self.return_value = json.dumps(data)
        self.data = data

    @property
    def foo_property(self):
        return self.data

    @foo_property.setter
    def foo_property(self, data):
        self.data = json.dumps(data)


def test_data_ingest_sanity(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )

    assert base_model._data_internal is base_model_data
    assert base_model._attributes_to_ignore is attributes_to_ignore
    assert base_model._attributes_to_remap is attributes_to_remap
    assert base_model._repr_metadata is repr_metadata


def test_getattr_name_in_attributes_to_remap(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )

    assert base_model.committed_by == base_model_data['committer']
    assert base_model.topology_name == base_model_data['name']


def test_getattr_python_to_json_attribute(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )
    assert base_model.topologyDefinition == base_model_data['topologyDefinition']
    assert base_model.topology_definition == base_model_data['topologyDefinition']


def test_setattr_python_to_json_attribute(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    data = {"foo": "baz"}
    helper_obj = Helper(data, base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata)
    assert helper_obj.foo_property == data  # Sanity check

    # Reassign value so property setter runs json.dumps()
    helper_obj.foo_property = data

    # Expect property setter to get called which returns a json.dumps() of data
    assert helper_obj.foo_property == helper_obj.return_value


def test_attributes_to_ignore_in_base_model(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    data = {"foo": "baz"}
    helper_obj = Helper(data, base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata)
    assert helper_obj._data["provenanceMetaData"] == base_model_data["provenanceMetaData"]  # Sanity check

    # Check if only the specified attribute is ignored
    assert hasattr(helper_obj, "topologyId")
    assert not hasattr(helper_obj, "provenanceMetaData")


def test_override_equal(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )
    copy_base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )

    assert base_model == copy_base_model


def test_copying_base_model(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    # create a base model that has another base model object as an attribute
    base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )
    sub_base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )
    base_model.sub_base_model = sub_base_model

    # copy to ensure base models are equal but not the same, sub_base_model should be equal and same
    copy_base_model = copy(base_model)
    assert base_model == copy_base_model
    assert id(base_model) != id(copy_base_model)
    assert base_model.sub_base_model == copy_base_model.sub_base_model
    assert id(base_model.sub_base_model) == id(copy_base_model.sub_base_model)


def test_deep_copying_base_model(base_model_data, attributes_to_ignore, attributes_to_remap, repr_metadata):
    # create a base model that has another base model object as an attribute
    base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )
    sub_base_model = BaseModel(
        data=base_model_data,
        attributes_to_ignore=attributes_to_ignore,
        attributes_to_remap=attributes_to_remap,
        repr_metadata=repr_metadata,
    )
    base_model.sub_base_model = sub_base_model

    # deepcopy to ensure both base_model and sub_base_model are equal to deepcopy, but their ids shouldn't be
    deepcopy_base_model = deepcopy(base_model)
    assert base_model == deepcopy_base_model
    assert id(base_model) != id(deepcopy_base_model)
    assert base_model.sub_base_model == deepcopy_base_model.sub_base_model
    assert id(base_model.sub_base_model) != id(deepcopy_base_model.sub_base_model)



# Copyright 2022 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.utils import get_random_string

# fmt: on

NUM_PIPELINES = 12
NUM_JOBS = 5


@pytest.fixture(scope="module")
def sample_pipelines(resources_label, sch, sch_authoring_sdc_id):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    pipelines = []
    try:
        for i in range(NUM_PIPELINES):
            pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
            dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
            trash = pipeline_builder.add_stage('Trash')
            dev_data_generator >> trash
            pipeline_name = 'simple_saql_pipeline_{}_{}_{}_{}'.format(
                'even' if i % 2 == 0 else 'odd', resources_label, i, get_random_string()
            )
            pipeline = pipeline_builder.build(pipeline_name)
            sch.publish_pipeline(pipeline)

            pipelines.append(pipeline)

        yield pipelines
    finally:
        for pipeline in pipelines:
            sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def sample_fragments(resources_label, sch, sch_authoring_sdc_id):
    """A set of trivial pipelines:

    dev_data_generator >> trash
    """
    pipelines = []
    try:
        for i in range(NUM_PIPELINES):
            pipeline_builder = sch.get_pipeline_builder(
                engine_type='data_collector', engine_id=sch_authoring_sdc_id, fragment=True
            )
            dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
            trash = pipeline_builder.add_stage('Trash')
            dev_data_generator >> trash
            pipeline_name = 'simple_saql_fragments_{}_{}_{}_{}'.format(
                'even' if i % 2 == 0 else 'odd', resources_label, i, get_random_string()
            )
            pipeline = pipeline_builder.build(pipeline_name)
            sch.publish_pipeline(pipeline)

            pipelines.append(pipeline)

        yield pipelines
    finally:
        for pipeline in pipelines:
            sch.delete_pipeline(pipeline)


@pytest.fixture(scope="module")
def simple_pipeline(resources_label, sch, sch_authoring_sdc_id):
    """A trivial pipeline:

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)

    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple_saql_draft_{}_{}'.format(resources_label, get_random_string()))
    sch.publish_pipeline(pipeline)

    try:
        yield pipeline
    finally:
        sch.api_client.delete_pipeline(pipeline.pipeline_id)


@pytest.fixture(scope="module")
def sample_jobs(resources_label, sch, simple_pipeline, sch_executor_sdc_label):
    """A set of simple jobs based on simple pipeline."""
    job_builder = sch.get_job_builder()

    jobs = []
    for i in range(NUM_JOBS):
        job = job_builder.build(
            'test_simple_saql_job_fetch_{}_{}_{}'.format(resources_label, i, get_random_string()),
            pipeline=simple_pipeline,
        )
        job.data_collector_labels = sch_executor_sdc_label
        sch.add_job(job)
        jobs.append(job)

    try:
        yield jobs
    finally:
        for job in jobs:
            sch.delete_job(job)


@pytest.fixture(scope="module")
def sample_job_templates(resources_label, sch, simple_pipeline, sch_executor_sdc_label):
    job_builder = sch.get_job_builder()

    jobs = []
    for i in range(NUM_JOBS):
        job = job_builder.build(
            'test_simple_saql_job_fetch_{}_{}_{}'.format(resources_label, i, get_random_string()),
            pipeline=simple_pipeline,
            job_template=True,
            runtime_parameters={'x': 'y', 'a': 'b'},
        )
        job.data_collector_labels = sch_executor_sdc_label
        sch.add_job(job)
        jobs.append(job)

    try:
        yield jobs
    finally:
        for job in jobs:
            sch.delete_job(job)


def test_saql_search_on_pipelines(resources_label, sch, sample_pipelines):
    even_pipelines = sch.pipelines.get_all(search="name==simple_saql_pipeline_even_{}*".format(resources_label))
    assert len(even_pipelines) == (NUM_PIPELINES / 2)
    odd_pipelines = sch.pipelines.get_all(search="name==simple_saql_pipeline_odd_{}*".format(resources_label))
    assert len(odd_pipelines) == (NUM_PIPELINES / 2)
    null_pipelines = sch.pipelines.get_all(search="name==DOES_NOT_EXIST")
    assert len(null_pipelines) == 0


def test_saql_search_on_fragments(resources_label, sch, sample_fragments):
    even_fragments = sch.pipelines.get_all(
        fragment=True, search="name==simple_saql_fragments_even_{}*".format(resources_label)
    )
    assert len(even_fragments) == (NUM_PIPELINES / 2)
    odd_fragments = sch.pipelines.get_all(
        fragment=True, search="name==simple_saql_fragments_odd_{}*".format(resources_label)
    )
    assert len(odd_fragments) == (NUM_PIPELINES / 2)
    null_fragments = sch.pipelines.get_all(fragment=True, search="name==DOES_NOT_EXIST")
    assert len(null_fragments) == 0


def test_saql_search_on_jobs(resources_label, sch, sample_jobs):
    simple_jobs = sch.jobs.get_all(search="name=='test_simple_saql_job_fetch_{}*'".format(resources_label))
    assert len(simple_jobs) == NUM_JOBS
    null_jobs = sch.jobs.get_all(search="name==DOES_NOT_EXIST")
    assert len(null_jobs) == 0


def test_saql_search_on_job_templates(resources_label, sch, sample_job_templates):
    simple_job_templates = sch.jobs.get_all(
        search="name=='test_simple_saql_job_fetch_{}*'".format(resources_label), job_template=True
    )
    assert len(simple_job_templates) == NUM_JOBS
    null_job_templates = sch.jobs.get_all(search="name==DOES_NOT_EXIST", job_template=True)
    assert len(null_job_templates) == 0



# Copyright 2024 StreamSets Inc.

# fmt: off
from copy import deepcopy

import pytest

from streamsets.sdk.sch_models import Job, JobSequence
from streamsets.sdk.utils import SeekableList

from .resources.job_sequence_step_data import JOB_JSON, JOB_SEQUENCE_WITH_STEPS_JSON, JOB_WITHIN_JOB_SEQUENCE_JSON

# fmt: on
START_TIME_2099 = 4200268800
END_TIME_2099 = 4200268800
START_TIME_2100 = 4237401600
END_TIME_2100 = 4237996800
UTC_TIME_Z0NE = ' UTC'
BASIC_CRON_TAB_MASK = '0/1 * 1/1 * ? *'
NUM_OF_STEPS = 4
NUM_OF_JOBS = 3


class MockControlHub:
    def __init__(self):
        self.api_client = MockApiClient()
        self.organization = 'DUMMY ORG'

    @property
    def jobs(self):
        return MockResponse('dummy')


class MockApiClient:
    def __init__(self):
        self.value = 'value'

    def mark_job_as_finished(self, value):
        return MockReturn(value)

    def get_job_sequence_log_history(self, value1, value2, value3):
        return MockReturn([value1, value2, value3])

    def add_step_jobs_to_job_sequence(self, value1, value2):
        return MockReturn({value1: value2})

    def update_job_sequence_steps(self, value):
        pass

    def update_steps_of_job_sequence(self, value1, value2):
        return MockReturn({'steps': value2})

    def get_jobs(self, job_ids):
        result = []
        for job_id in job_ids:
            json_copy = dict(JOB_JSON)
            json_copy['id'] = job_id
            result.append(json_copy)
        return MockReturn(result)


class MockReturn:
    def __init__(self, value):
        self.response = MockResponse(value)


class MockResponse:
    def __init__(self, value):
        self.value = value

    def json(self):
        return self.value

    def get(self, job_id):
        json_copy = dict(JOB_JSON)
        json_copy['id'] = job_id
        job = Job(json_copy, MockControlHub())
        return job


@pytest.fixture(scope="function")
def job_sequence_with_steps():
    json = deepcopy(JOB_SEQUENCE_WITH_STEPS_JSON)
    return JobSequence(json, MockControlHub())


@pytest.fixture(scope="function")
def job_within_job_sequence():
    json = deepcopy(JOB_WITHIN_JOB_SEQUENCE_JSON)
    return Job(json, MockControlHub())


@pytest.fixture(scope="function")
def dummy_jobs():
    jobs = []
    for i in range(NUM_OF_JOBS):
        json_copy = dict(JOB_JSON)
        json_copy['id'] = i
        jobs.append(Job(json_copy, MockControlHub()))

    return jobs


def test_step_jobs(job_sequence_with_steps):
    step = job_sequence_with_steps.steps[0]

    assert isinstance(step.step_jobs, SeekableList)
    assert len(step.step_jobs) == 1
    assert isinstance(step.step_jobs[0], Job)
    assert step._data['jobs'][0]['jobId'] == step.step_jobs[0].job_id


def test_remove_jobs(job_sequence_with_steps):
    step = job_sequence_with_steps.steps[0]
    step_jobs = step.step_jobs

    assert len(step_jobs) == 1
    id = step.id
    job_to_remove = step_jobs[0]

    step.remove_jobs(job_to_remove)

    assert step._data['id'] == id
    assert len(step._data['jobIds']) == 0


def test_remove_jobs_invalid_type(job_sequence_with_steps):
    step = job_sequence_with_steps.steps[0]

    with pytest.raises(TypeError):
        step.remove_jobs('invalid string')


def test_remove_jobs_with_job_that_isnt_in_step(job_sequence_with_steps):
    step1, step2 = job_sequence_with_steps.steps[0], job_sequence_with_steps.steps[1]

    with pytest.raises(ValueError):
        step1.remove_jobs(step2.step_jobs[0])


def test_add_job_incorrect_jobs_type(job_sequence_with_steps):
    step = job_sequence_with_steps.steps[0]

    with pytest.raises(TypeError):
        step.add_jobs('1')

    with pytest.raises(TypeError):
        step.add_jobs(['1'])


def test_add_job_incorrect_ignore_error_type(job_sequence_with_steps, dummy_jobs):
    step = job_sequence_with_steps.steps[0]
    job = dummy_jobs[0]
    job._data['jobSequence'] = None

    with pytest.raises(TypeError):
        step.add_jobs([job], ignore_error='1')


def test_add_job_with_job_that_is_in_another_job_sequence(job_sequence_with_steps, dummy_jobs):
    step = job_sequence_with_steps.steps[0]
    job = dummy_jobs[0]
    job._data['jobSequence'] = job_sequence_with_steps

    with pytest.raises(ValueError):
        step.add_jobs([job])


def test_change_step_name(job_sequence_with_steps):
    step = job_sequence_with_steps.steps[0]
    assert step.name == 'Step 1'

    step.name = 'My new step'
    assert step.name == 'My new step'



# Copyright 2023 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.sch_models import BaseModel, CollectionModel, CollectionModelResults, MutableKwargs

# fmt: on


class ErroneousCollectionModel(CollectionModel):
    """
    A subclass of CollectionModel that does not implement the _get_all_results_from_api function
    """

    pass


class ValidCollectionModel(CollectionModel):
    """
    A subclass of CollectionModel that implements _get_all_results_from_api function
    """

    def __init__(self, control_hub):
        super().__init__(control_hub)
        self.data = [{'id': i} for i in range(100)]
        self.page_size = None

    def _get_all_results_from_api(self, **kwargs):
        kwargs_defaults = {'offset': None, 'len': None}
        kwargs_instance = MutableKwargs(kwargs_defaults, kwargs)
        kwargs_unioned = kwargs_instance.union()

        response = {
            'data': self.data[kwargs_unioned['offset'] : kwargs_unioned['offset'] + kwargs_unioned['len']],
            'offset': kwargs_unioned['offset'] + kwargs_unioned['len'],
            'len': kwargs_unioned['len'],
        }

        self.page_size = kwargs_unioned['len']
        kwargs_unused = kwargs_instance.subtract()

        return CollectionModelResults(response, kwargs_unused, BaseModel, {})

    def __len__(self):
        return len(self.data)


class DecreasingLengthCollectionModel(CollectionModel):
    """
    This class is used to test a collection model whose length decreases a lot suddenly

    What we want to test specifically:
    1. Initially we have 150 items
    2. user makes a call to get all items
    3. Since collection model gets minimum of (50, requested_length) we should get 50 items in the initial response
    4. We set the total number of items we have to a total of 50 (deleting 100 items)
    5. We raise an exception if offset is negative in _get_all_results_from_api and if no error is thrown we good.

    If repeated number of calls are made then we only have 50 data points to deal with
    """

    def __init__(self, control_hub):
        super().__init__(control_hub)
        self.data = [{'id': i} for i in range(150)]

    def _get_all_results_from_api(self, **kwargs):
        kwargs_defaults = {'offset': None, 'len': None}
        kwargs_instance = MutableKwargs(kwargs_defaults, kwargs)
        kwargs_unioned = kwargs_instance.union()

        if kwargs_unioned['offset'] < 0:
            raise ValueError("Got negative offset in _get_all_results_from_api")

        response = {
            'data': self.data[kwargs_unioned['offset'] : kwargs_unioned['offset'] + kwargs_unioned['len']],
            'offset': kwargs_unioned['offset'] + kwargs_unioned['len'],
            'len': 0,
        }  # len isn't used but is checked for
        kwargs_unused = kwargs_instance.subtract()

        # setting the length of items to 50, we do -50 because we have already seen the first few
        self.data = self.data[-50:]

        return CollectionModelResults(response, kwargs_unused, BaseModel, {})

    def __len__(self):
        return len(self.data)


def test_not_implemented_error(mocker):
    invalid_collection_model = ErroneousCollectionModel(mocker.Mock())
    with pytest.raises(NotImplementedError):
        invalid_collection_model._get_all_results_from_api()


def test_invalid_offset(mocker):
    valid_collection_model = ValidCollectionModel(mocker.Mock())
    with pytest.raises(ValueError):
        next(valid_collection_model._paginate(offset=-1))  # _paginate is a generator


@pytest.mark.parametrize("offset", [0, 10])
def test_valid_offset(mocker, offset):
    valid_collection_model = ValidCollectionModel(mocker.Mock())

    res = valid_collection_model._paginate(offset=offset)
    res = list(res)

    expected_result_length = len(valid_collection_model.data) - offset
    assert len(res) == expected_result_length


def test_valid_page_size_is_250(mocker):
    valid_collection_model = ValidCollectionModel(mocker.Mock())
    list(valid_collection_model._paginate())
    assert valid_collection_model.page_size == CollectionModel.PAGE_LENGTH


def test_decreasing_length_collection_model(mocker):
    decreasing_collection_model = DecreasingLengthCollectionModel(mocker.Mock())
    res = decreasing_collection_model._paginate(offset=0, len=75)  # we want more items than the page length

    assert len(list(res)) == 75



# Copyright 2024 StreamSets Inc.


def test_deployment_acl(sch, sdc_deployment):
    NEW_ACTION = ["READ"]
    deployment_acl = sdc_deployment.acl
    permission = deployment_acl.permissions[0]
    original_actions = permission.actions
    try:
        assert permission.actions != NEW_ACTION
        permission.actions = NEW_ACTION
        assert permission.actions == NEW_ACTION
        deployment_acl.add_permission(permission)
        assert deployment_acl._data["permissions"][0]["actions"] == NEW_ACTION
        # Double check that the deployment was actually updated
        deployment = sch.deployments.get(deployment_id=sdc_deployment.deployment_id)
        assert deployment.acl._data["permissions"][0]["actions"] == NEW_ACTION
    finally:
        permission.actions = original_actions
        deployment_acl.add_permission(permission)


def test_environment_acl(sch, test_environment):
    NEW_ACTION = ["READ"]
    environment_acl = test_environment.acl
    permission = environment_acl.permissions[0]
    original_actions = permission.actions
    try:
        assert permission.actions != NEW_ACTION
        permission.actions = NEW_ACTION
        assert permission.actions == NEW_ACTION
        environment_acl.add_permission(permission)
        assert environment_acl._data["permissions"][0]["actions"] == NEW_ACTION
        # Double check that the deployment was actually updated
        environment = sch.environments.get(environment_id=test_environment.environment_id)
        assert environment.acl._data["permissions"][0]["actions"] == NEW_ACTION
    finally:
        permission.actions = original_actions
        environment_acl.add_permission(permission)



JOB_SEQUENCE_BUILDER_JSON = {
    'type': 'object',
    'properties': {
        'name': {'type': 'string'},
        'description': {'type': 'string'},
        'startTimeEnabled': {'type': 'boolean'},
        'startTime': {'type': 'integer', 'format': 'int64'},
        'repeatEnabled': {'type': 'boolean'},
        'crontabMask': {'type': 'string'},
        'endTimeCondition': {'type': 'string', 'enum': ['NEVER', 'ON']},
        'endTime': {'type': 'integer', 'format': 'int64'},
        'timezone': {'type': 'string'},
        'status': {'type': 'string', 'enum': ['INACTIVE', 'ACTIVE', 'DISABLED', 'ERROR']},
    },
}

JOB_SEQUENCE_EMPTY_JSON = {
    "id": None,
    "name": None,
    "description": None,
    "startTime": None,
    "endTime": None,
    "timezone": None,
    "crontabMask": None,
    "status": None,
    "startTimeEnabled": None,
    "repeatEnabled": None,
    "endTimeCondition": None,
    "nextRunTime": None,
}

PIPELINE_BUILDER_JSON = {
    "pipelineId": 1,
    "commitId": 1,
    "name": "Test Pipeline",
    "version": 1,
    "sdcId": "f0c68f66-9bf9-41cb-869a-2f703970d356",
    "pipelineConfig": {
        "schemaVersion": 6,
        "version": 39,
        "pipelineId": None,
        "title": "Pipeline Test",
        "description": "",
        "uuid": "13ee6ba6-f738-459c-b7e1-0002f2d13248",
        "configuration": [],
        "uiInfo": {},
        "info": {},
        "fragments": [],
        "stages": [],
        "errorStage": {
            "instanceName": "errorStageStageInstance",
            "library": "streamsets-datacollector-basic-lib",
            "stageName": "com_streamsets_pipeline_stage_destination_devnull_ToErrorNullDTarget",
            "stageVersion": "1",
            "configuration": [],
            "uiInfo": {"stageType": "TARGET", "label": "Error -Discard"},
            "inputLanes": [],
            "outputLanes": [],
            "eventLanes": [],
            "services": [],
        },
    },
}


PIPELINE_BUILDER_DEFINITIONS_JSON = {
    'stages': [
        {  # snowflake destination
            "services": [],
            "description": "Fast data upload to Snowflake with Data Drift and CDC support",
            "name": "com_streamsets_pipeline_stage_destination_snowflake_SnowflakeDTarget",
            "type": "TARGET",
            "className": "com.streamsets.pipeline.stage.destination.snowflake.SnowflakeDTarget",
            "label": "Snowflake",
            "version": "18",
            "library": "streamsets-datacollector-sdc-snowflake-lib",
            "libraryLabel": "Snowflake Library",
            "configDefinitions": [
                {
                    "fieldName": "snowflakeConnectionSelection",
                    "description": "",
                    "name": "config.snowflake.snowflakeConnectionSelection",
                    "type": "MODEL",
                    "defaultValue": "MANUAL",
                    "label": "Connection",
                    "connectionType": "STREAMSETS_SNOWFLAKE",
                    "model": {
                        "labels": ["None"],
                        "values": ["MANUAL"],
                        "configDefinitions": None,
                        "modelType": "VALUE_CHOOSER",
                        "compositeConfigDefinitions": None,
                        "valuesProviderClass": "com.streamsets.pipeline.api."
                        "ConnectionDef$Constants$ConnectionChooserValues",
                        "filteringConfig": "",
                    },
                }
            ],
            "icon": "snowflake.png",
            "errorStage": False,
            "outputStreamLabels": None,
            "variableOutputStreams": False,
            "outputStreams": 0,
            "outputStreamLabelProviderClass": None,
            "inputStreamLabels": None,
            "inputStreams": -1,
            "inputStreamLabelProviderClass": "com.streamsets.pipeline.api.StageDef$DefaultInputStreams",
            "resetOffset": False,
            "outputStreamsDrivenByConfig": "",
            "producingEvents": False,
            "preconditions": True,
            "statsAggregatorStage": False,
        },
        {
            'services': [],
            'label': 'Trash',
            'name': 'com_streamsets_pipeline_stage_destination_devnull_NullDTarget',
            'type': 'TARGET',
            'className': 'com.streamsets.pipeline.stage.destination.devnull.NullDTarget',
            'version': '1',
            'eventDefs': [],
            'library': 'streamsets-datacollector-basic-lib',
            'privateClassLoader': False,
            'libraryLabel': 'Basic',
            'configDefinitions': [],
            'explorerSchema': None,
            'configGroupDefinition': {'classNameToGroupsMap': {}, 'groupNameToLabelMapList': []},
            'compositeConfigDefinitions': [],
            'icon': 'trash.png',
            'pipelineLifecycleStage': False,
            'statsAggregatorStage': False,
            'errorStage': False,
            'rawSourceDefinition': None,
            'outputStreamLabels': None,
            'variableOutputStreams': False,
            'outputStreams': 0,
            'outputStreamLabelProviderClass': None,
            'inputStreamLabels': None,
            'inputStreams': -1,
            'inputStreamLabelProviderClass': 'com.streamsets.pipeline.api.StageDef$DefaultInputStreams',
            'connectionVerifierStage': False,
            'offsetCommitTrigger': False,
            'executionModes': [
                'STANDALONE',
                'CLUSTER_BATCH',
                'CLUSTER_YARN_STREAMING',
                'CLUSTER_MESOS_STREAMING',
                'EDGE',
                'EMR_BATCH',
            ],
            'resetOffset': False,
        },
        {
            'label': 'Dev Data Generator',
            'services': [],
            'description': 'Generates records with the specified field names based on the selected data type. For development only.',
            'name': 'com_streamsets_pipeline_stage_devtest_RandomDataGeneratorSource',
            'type': 'SOURCE',
            'className': 'com.streamsets.pipeline.stage.devtest.RandomDataGeneratorSource',
            'version': '6',
            'pipelineLifecycleStage': False,
            'eventDefs': [],
            'library': 'streamsets-datacollector-dev-lib',
            'privateClassLoader': False,
            'libraryLabel': 'Dev (for development only)',
            'configDefinitions': [
                {
                    'fieldName': 'delay',
                    'max': 2147483647,
                    'label': 'Delay Between Batches',
                    'description': 'Milliseconds to wait before sending the next batch',
                    'name': 'delay',
                    'type': 'NUMBER',
                    'defaultValue': 1000,
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': None,
                    'required': True,
                    'min': 0,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'numberOfRecords',
                    'max': 2147483647,
                    'label': 'Records to be Generated',
                    'description': 'Number of records to be generated. 0 means infinite',
                    'name': 'numberOfRecords',
                    'type': 'NUMBER',
                    'defaultValue': 0,
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': None,
                    'required': True,
                    'min': 0,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'eventName',
                    'max': 9223372036854775807,
                    'label': 'Event Name',
                    'description': 'Name of event that should be used when generating events.',
                    'name': 'eventName',
                    'type': 'STRING',
                    'defaultValue': 'generated-event',
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': None,
                    'required': True,
                    'min': -9223372036854775808,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'batchSize',
                    'max': 2147483647,
                    'label': 'Batch Size',
                    'description': 'Number of records that will be generated for single batch.',
                    'name': 'batchSize',
                    'type': 'NUMBER',
                    'defaultValue': 1000,
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': None,
                    'required': True,
                    'min': 1,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'headerAttributes',
                    'max': 9223372036854775807,
                    'label': 'Header Attributes',
                    'description': 'Attributes to be put in the generated record header',
                    'name': 'headerAttributes',
                    'type': 'MAP',
                    'defaultValue': None,
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': None,
                    'required': False,
                    'min': -9223372036854775808,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'BASIC',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'dataGenConfigs',
                    'max': 9223372036854775807,
                    'label': 'Fields to Generate',
                    'description': 'Fields to generate of the indicated type',
                    'name': 'dataGenConfigs',
                    'type': 'MODEL',
                    'defaultValue': None,
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': {
                        'labels': None,
                        'values': None,
                        'configDefinitions': [
                            {
                                'fieldName': 'field',
                                'max': 9223372036854775807,
                                'label': 'Field Name',
                                'description': '',
                                'name': 'field',
                                'type': 'STRING',
                                'defaultValue': None,
                                'lines': 0,
                                'mode': 'text/plain',
                                'model': None,
                                'required': True,
                                'min': -9223372036854775808,
                                'dependsOnMap': {},
                                'dependsOn': '',
                                'explorerConfig': None,
                                'group': '',
                                'evaluation': 'IMPLICIT',
                                'elDefs': None,
                                'upload': 'NO',
                                'displayMode': 'BASIC',
                                'triggeredByValues': None,
                                'displayPosition': 0,
                                'connectionType': '',
                                'requiredForExplorerLoad': False,
                                'placeholder': {'text': ''},
                                'elFunctionDefinitionsIdx': [
                                    '1',
                                    '2',
                                    '3',
                                    '4',
                                    '5',
                                    '6',
                                    '7',
                                    '8',
                                    '9',
                                    '10',
                                    '11',
                                    '12',
                                    '13',
                                    '14',
                                    '15',
                                    '16',
                                    '17',
                                    '18',
                                    '19',
                                    '20',
                                    '21',
                                    '22',
                                    '23',
                                    '24',
                                    '25',
                                    '26',
                                    '27',
                                    '28',
                                    '29',
                                    '30',
                                    '31',
                                    '32',
                                    '33',
                                    '34',
                                    '35',
                                    '36',
                                    '37',
                                    '38',
                                    '39',
                                    '40',
                                    '41',
                                    '42',
                                    '43',
                                    '44',
                                    '45',
                                    '46',
                                    '47',
                                    '48',
                                    '49',
                                    '50',
                                    '51',
                                    '52',
                                    '53',
                                    '54',
                                    '55',
                                    '56',
                                    '57',
                                    '58',
                                    '59',
                                    '60',
                                    '61',
                                    '62',
                                    '63',
                                    '64',
                                ],
                                'elConstantDefinitionsIdx': ['65'],
                                'constraints': [],
                            },
                            {
                                'fieldName': 'type',
                                'max': 9223372036854775807,
                                'label': 'Field Type',
                                'description': '',
                                'name': 'type',
                                'type': 'MODEL',
                                'defaultValue': 'STRING',
                                'lines': 0,
                                'mode': 'text/plain',
                                'model': {
                                    'labels': [
                                        'STRING',
                                        'INTEGER',
                                        'LONG',
                                        'FLOAT',
                                        'DOUBLE',
                                        'DATE',
                                        'DATETIME',
                                        'ZONED_DATETIME',
                                        'TIME',
                                        'BOOLEAN',
                                        'DECIMAL',
                                        'BYTE_ARRAY',
                                        'LONG_SEQUENCE',
                                        'ADDRESS_FULL_ADDRESS',
                                        'ADDRESS_BUILDING_NUMBER',
                                        'ADDRESS_STREET_ADDRESS',
                                        'ADDRESS_CITY',
                                        'ADDRESS_STATE',
                                        'ADDRESS_COUNTRY',
                                        'ADDRESS_LATITUDE',
                                        'ADDRESS_LONGITUDE',
                                        'APP_NAME',
                                        'APP_AUTHOR',
                                        'APP_VERSION',
                                        'ARTIST_NAME',
                                        'BOOK_TITLE',
                                        'BOOK_AUTHOR',
                                        'BOOK_GENRE',
                                        'BOOK_PUBLISHER',
                                        'BUSINESS_CREDIT_CARD_NUMBER',
                                        'BUSINESS_CREDIT_CARD_EXPIRY',
                                        'BUSINESS_CREDIT_CARD_TYPE',
                                        'CAT_NAME',
                                        'CAT_BREED',
                                        'CAT_REGISTRY',
                                        'CODE_ASIN',
                                        'CODE_IMEI',
                                        'CODE_ISBN10',
                                        'CODE_ISBN13',
                                        'COLOR',
                                        'COMMERCE_DEPARTMENT',
                                        'COMMERCE_COLOR',
                                        'COMMERCE_MATERIAL',
                                        'COMMERCE_PRICE',
                                        'COMMERCE_PRODUCT_NAME',
                                        'COMMERCE_PROMOTION_CODE',
                                        'COMPANY_NAME',
                                        'COMPANY_INDUSTRY',
                                        'COMPANY_BUZZWORD',
                                        'COMPANY_URL',
                                        'CRYPTO_MD5',
                                        'CRYPTO_SHA1',
                                        'CRYPTO_SHA256',
                                        'CRYPTO_SHA512',
                                        'DEMOGRAPHIC',
                                        'EDUCATOR',
                                        'EMAIL',
                                        'FILE',
                                        'FINANCE',
                                        'FOOD',
                                        'GAMEOFTHRONES',
                                        'HACKER',
                                        'IDNUMBER',
                                        'INTERNET',
                                        'LOREM',
                                        'MUSIC',
                                        'NAME',
                                        'PHONENUMBER',
                                        'POKEMON',
                                        'RACE',
                                        'SEX',
                                        'SHAKESPEARE',
                                        'SLACKEMOJI',
                                        'SPACE',
                                        'SSN',
                                        'STOCK',
                                        'SUPERHERO',
                                        'TEAM',
                                        'UNIVERSITY',
                                    ],
                                    'values': [
                                        'STRING',
                                        'INTEGER',
                                        'LONG',
                                        'FLOAT',
                                        'DOUBLE',
                                        'DATE',
                                        'DATETIME',
                                        'ZONED_DATETIME',
                                        'TIME',
                                        'BOOLEAN',
                                        'DECIMAL',
                                        'BYTE_ARRAY',
                                        'LONG_SEQUENCE',
                                        'ADDRESS_FULL_ADDRESS',
                                        'ADDRESS_BUILDING_NUMBER',
                                        'ADDRESS_STREET_ADDRESS',
                                        'ADDRESS_CITY',
                                        'ADDRESS_STATE',
                                        'ADDRESS_COUNTRY',
                                        'ADDRESS_LATITUDE',
                                        'ADDRESS_LONGITUDE',
                                        'APP_NAME',
                                        'APP_AUTHOR',
                                        'APP_VERSION',
                                        'ARTIST_NAME',
                                        'BOOK_TITLE',
                                        'BOOK_AUTHOR',
                                        'BOOK_GENRE',
                                        'BOOK_PUBLISHER',
                                        'BUSINESS_CREDIT_CARD_NUMBER',
                                        'BUSINESS_CREDIT_CARD_EXPIRY',
                                        'BUSINESS_CREDIT_CARD_TYPE',
                                        'CAT_NAME',
                                        'CAT_BREED',
                                        'CAT_REGISTRY',
                                        'CODE_ASIN',
                                        'CODE_IMEI',
                                        'CODE_ISBN10',
                                        'CODE_ISBN13',
                                        'COLOR',
                                        'COMMERCE_DEPARTMENT',
                                        'COMMERCE_COLOR',
                                        'COMMERCE_MATERIAL',
                                        'COMMERCE_PRICE',
                                        'COMMERCE_PRODUCT_NAME',
                                        'COMMERCE_PROMOTION_CODE',
                                        'COMPANY_NAME',
                                        'COMPANY_INDUSTRY',
                                        'COMPANY_BUZZWORD',
                                        'COMPANY_URL',
                                        'CRYPTO_MD5',
                                        'CRYPTO_SHA1',
                                        'CRYPTO_SHA256',
                                        'CRYPTO_SHA512',
                                        'DEMOGRAPHIC',
                                        'EDUCATOR',
                                        'EMAIL',
                                        'FILE',
                                        'FINANCE',
                                        'FOOD',
                                        'GAMEOFTHRONES',
                                        'HACKER',
                                        'IDNUMBER',
                                        'INTERNET',
                                        'LOREM',
                                        'MUSIC',
                                        'NAME',
                                        'PHONENUMBER',
                                        'POKEMON',
                                        'RACE',
                                        'SEX',
                                        'SHAKESPEARE',
                                        'SLACKEMOJI',
                                        'SPACE',
                                        'SSN',
                                        'STOCK',
                                        'SUPERHERO',
                                        'TEAM',
                                        'UNIVERSITY',
                                    ],
                                    'configDefinitions': None,
                                    'modelType': 'VALUE_CHOOSER',
                                    'compositeConfigDefinitions': None,
                                    'filteringConfig': '',
                                    'valuesProviderClass': 'com.streamsets.pipeline.stage.devtest.TypeChooserValueProvider',
                                },
                                'required': True,
                                'min': -9223372036854775808,
                                'dependsOnMap': {},
                                'dependsOn': '',
                                'explorerConfig': None,
                                'group': '',
                                'evaluation': 'IMPLICIT',
                                'elDefs': None,
                                'upload': 'NO',
                                'displayMode': 'BASIC',
                                'triggeredByValues': None,
                                'displayPosition': 0,
                                'connectionType': '',
                                'requiredForExplorerLoad': False,
                                'placeholder': {'text': ''},
                                'elFunctionDefinitionsIdx': [
                                    '1',
                                    '2',
                                    '3',
                                    '4',
                                    '5',
                                    '6',
                                    '7',
                                    '8',
                                    '9',
                                    '10',
                                    '11',
                                    '12',
                                    '13',
                                    '14',
                                    '15',
                                    '16',
                                    '17',
                                    '18',
                                    '19',
                                    '20',
                                    '21',
                                    '22',
                                    '23',
                                    '24',
                                    '25',
                                    '26',
                                    '27',
                                    '28',
                                    '29',
                                    '30',
                                    '31',
                                    '32',
                                    '33',
                                    '34',
                                    '35',
                                    '36',
                                    '37',
                                    '38',
                                    '39',
                                    '40',
                                    '41',
                                    '42',
                                    '43',
                                    '44',
                                    '45',
                                    '46',
                                    '47',
                                    '48',
                                    '49',
                                    '50',
                                    '51',
                                    '52',
                                    '53',
                                    '54',
                                    '55',
                                    '56',
                                    '57',
                                    '58',
                                    '59',
                                    '60',
                                    '61',
                                    '62',
                                    '63',
                                    '64',
                                ],
                                'elConstantDefinitionsIdx': ['65'],
                                'constraints': [],
                            },
                            {
                                'fieldName': 'precision',
                                'max': 9223372036854775807,
                                'label': 'Precision',
                                'description': 'Precision of the generated decimal.',
                                'name': 'precision',
                                'type': 'NUMBER',
                                'defaultValue': 10,
                                'lines': 0,
                                'mode': 'text/plain',
                                'model': None,
                                'required': True,
                                'min': 0,
                                'dependsOnMap': {'type': ['DECIMAL']},
                                'dependsOn': '',
                                'explorerConfig': None,
                                'group': '',
                                'evaluation': 'IMPLICIT',
                                'elDefs': None,
                                'upload': 'NO',
                                'displayMode': 'BASIC',
                                'triggeredByValues': None,
                                'displayPosition': 0,
                                'connectionType': '',
                                'requiredForExplorerLoad': False,
                                'placeholder': {'text': ''},
                                'elFunctionDefinitionsIdx': [
                                    '1',
                                    '2',
                                    '3',
                                    '4',
                                    '5',
                                    '6',
                                    '7',
                                    '8',
                                    '9',
                                    '10',
                                    '11',
                                    '12',
                                    '13',
                                    '14',
                                    '15',
                                    '16',
                                    '17',
                                    '18',
                                    '19',
                                    '20',
                                    '21',
                                    '22',
                                    '23',
                                    '24',
                                    '25',
                                    '26',
                                    '27',
                                    '28',
                                    '29',
                                    '30',
                                    '31',
                                    '32',
                                    '33',
                                    '34',
                                    '35',
                                    '36',
                                    '37',
                                    '38',
                                    '39',
                                    '40',
                                    '41',
                                    '42',
                                    '43',
                                    '44',
                                    '45',
                                    '46',
                                    '47',
                                    '48',
                                    '49',
                                    '50',
                                    '51',
                                    '52',
                                    '53',
                                    '54',
                                    '55',
                                    '56',
                                    '57',
                                    '58',
                                    '59',
                                    '60',
                                    '61',
                                    '62',
                                    '63',
                                    '64',
                                ],
                                'elConstantDefinitionsIdx': ['65'],
                                'constraints': [],
                            },
                            {
                                'fieldName': 'scale',
                                'max': 9223372036854775807,
                                'label': 'Scale',
                                'description': 'Scale of the generated decimal.',
                                'name': 'scale',
                                'type': 'NUMBER',
                                'defaultValue': 2,
                                'lines': 0,
                                'mode': 'text/plain',
                                'model': None,
                                'required': True,
                                'min': 0,
                                'dependsOnMap': {'type': ['DECIMAL']},
                                'dependsOn': '',
                                'explorerConfig': None,
                                'group': '',
                                'evaluation': 'IMPLICIT',
                                'elDefs': None,
                                'upload': 'NO',
                                'displayMode': 'BASIC',
                                'triggeredByValues': None,
                                'displayPosition': 0,
                                'connectionType': '',
                                'requiredForExplorerLoad': False,
                                'placeholder': {'text': ''},
                                'elFunctionDefinitionsIdx': [
                                    '1',
                                    '2',
                                    '3',
                                    '4',
                                    '5',
                                    '6',
                                    '7',
                                    '8',
                                    '9',
                                    '10',
                                    '11',
                                    '12',
                                    '13',
                                    '14',
                                    '15',
                                    '16',
                                    '17',
                                    '18',
                                    '19',
                                    '20',
                                    '21',
                                    '22',
                                    '23',
                                    '24',
                                    '25',
                                    '26',
                                    '27',
                                    '28',
                                    '29',
                                    '30',
                                    '31',
                                    '32',
                                    '33',
                                    '34',
                                    '35',
                                    '36',
                                    '37',
                                    '38',
                                    '39',
                                    '40',
                                    '41',
                                    '42',
                                    '43',
                                    '44',
                                    '45',
                                    '46',
                                    '47',
                                    '48',
                                    '49',
                                    '50',
                                    '51',
                                    '52',
                                    '53',
                                    '54',
                                    '55',
                                    '56',
                                    '57',
                                    '58',
                                    '59',
                                    '60',
                                    '61',
                                    '62',
                                    '63',
                                    '64',
                                ],
                                'elConstantDefinitionsIdx': ['65'],
                                'constraints': [],
                            },
                            {
                                'fieldName': 'fieldAttributes',
                                'max': 9223372036854775807,
                                'label': 'Field Attributes',
                                'description': 'Attributes to add to the generated field',
                                'name': 'fieldAttributes',
                                'type': 'MAP',
                                'defaultValue': None,
                                'lines': 0,
                                'mode': 'text/plain',
                                'model': None,
                                'required': False,
                                'min': -9223372036854775808,
                                'dependsOnMap': {},
                                'dependsOn': '',
                                'explorerConfig': None,
                                'group': '',
                                'evaluation': 'IMPLICIT',
                                'elDefs': None,
                                'upload': 'NO',
                                'displayMode': 'BASIC',
                                'triggeredByValues': None,
                                'displayPosition': 0,
                                'connectionType': '',
                                'requiredForExplorerLoad': False,
                                'placeholder': {'text': ''},
                                'elFunctionDefinitionsIdx': [
                                    '1',
                                    '2',
                                    '3',
                                    '4',
                                    '5',
                                    '6',
                                    '7',
                                    '8',
                                    '9',
                                    '10',
                                    '11',
                                    '12',
                                    '13',
                                    '14',
                                    '15',
                                    '16',
                                    '17',
                                    '18',
                                    '19',
                                    '20',
                                    '21',
                                    '22',
                                    '23',
                                    '24',
                                    '25',
                                    '26',
                                    '27',
                                    '28',
                                    '29',
                                    '30',
                                    '31',
                                    '32',
                                    '33',
                                    '34',
                                    '35',
                                    '36',
                                    '37',
                                    '38',
                                    '39',
                                    '40',
                                    '41',
                                    '42',
                                    '43',
                                    '44',
                                    '45',
                                    '46',
                                    '47',
                                    '48',
                                    '49',
                                    '50',
                                    '51',
                                    '52',
                                    '53',
                                    '54',
                                    '55',
                                    '56',
                                    '57',
                                    '58',
                                    '59',
                                    '60',
                                    '61',
                                    '62',
                                    '63',
                                    '64',
                                ],
                                'elConstantDefinitionsIdx': ['65'],
                                'constraints': [],
                            },
                        ],
                        'modelType': 'LIST_BEAN',
                        'compositeConfigDefinitions': [],
                        'filteringConfig': None,
                        'valuesProviderClass': None,
                    },
                    'required': False,
                    'min': -9223372036854775808,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'BASIC',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [],
                    'elConstantDefinitionsIdx': [],
                    'constraints': [],
                },
                {
                    'fieldName': 'numThreads',
                    'max': 2147483647,
                    'label': 'Number of Threads',
                    'description': 'Number of concurrent threads that will be generating data in parallel.',
                    'name': 'numThreads',
                    'type': 'NUMBER',
                    'defaultValue': 1,
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': None,
                    'required': True,
                    'min': 1,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'rootFieldType',
                    'max': 9223372036854775807,
                    'label': 'Root Field Type',
                    'description': 'Field Type for root object',
                    'name': 'rootFieldType',
                    'type': 'MODEL',
                    'defaultValue': 'MAP',
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': {
                        'labels': ['MAP', 'LIST_MAP'],
                        'values': ['MAP', 'LIST_MAP'],
                        'configDefinitions': None,
                        'modelType': 'VALUE_CHOOSER',
                        'compositeConfigDefinitions': None,
                        'filteringConfig': '',
                        'valuesProviderClass': 'com.streamsets.pipeline.stage.devtest.RootTypeChooserValueProvider',
                    },
                    'required': True,
                    'min': -9223372036854775808,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': 'DATA_GENERATOR',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 0,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
                {
                    'fieldName': 'stageOnRecordError',
                    'max': 9223372036854775807,
                    'label': 'On Record Error',
                    'description': 'Action to take with records sent to error',
                    'name': 'stageOnRecordError',
                    'type': 'MODEL',
                    'defaultValue': 'TO_ERROR',
                    'lines': 0,
                    'mode': 'text/plain',
                    'model': {
                        'labels': ['Discard', 'Send to Error', 'Stop Pipeline'],
                        'values': ['DISCARD', 'TO_ERROR', 'STOP_PIPELINE'],
                        'configDefinitions': None,
                        'modelType': 'VALUE_CHOOSER',
                        'compositeConfigDefinitions': None,
                        'filteringConfig': '',
                        'valuesProviderClass': 'com.streamsets.datacollector.config.OnRecordErrorChooserValues',
                    },
                    'required': True,
                    'min': -9223372036854775808,
                    'dependsOnMap': {},
                    'dependsOn': '',
                    'explorerConfig': None,
                    'group': '',
                    'evaluation': 'IMPLICIT',
                    'elDefs': None,
                    'upload': 'NO',
                    'displayMode': 'ADVANCED',
                    'triggeredByValues': None,
                    'displayPosition': 30,
                    'connectionType': '',
                    'requiredForExplorerLoad': False,
                    'placeholder': {'text': ''},
                    'elFunctionDefinitionsIdx': [
                        '1',
                        '2',
                        '3',
                        '4',
                        '5',
                        '6',
                        '7',
                        '8',
                        '9',
                        '10',
                        '11',
                        '12',
                        '13',
                        '14',
                        '15',
                        '16',
                        '17',
                        '18',
                        '19',
                        '20',
                        '21',
                        '22',
                        '23',
                        '24',
                        '25',
                        '26',
                        '27',
                        '28',
                        '29',
                        '30',
                        '31',
                        '32',
                        '33',
                        '34',
                        '35',
                        '36',
                        '37',
                        '38',
                        '39',
                        '40',
                        '41',
                        '42',
                        '43',
                        '44',
                        '45',
                        '46',
                        '47',
                        '48',
                        '49',
                        '50',
                        '51',
                        '52',
                        '53',
                        '54',
                        '55',
                        '56',
                        '57',
                        '58',
                        '59',
                        '60',
                        '61',
                        '62',
                        '63',
                        '64',
                    ],
                    'elConstantDefinitionsIdx': ['65'],
                    'constraints': [],
                },
            ],
            'icon': 'dev.png',
            'statsAggregatorStage': False,
            'errorStage': False,
            'rawSourceDefinition': None,
            'outputStreamLabels': ['Output'],
            'variableOutputStreams': False,
            'outputStreams': 1,
            'outputStreamLabelProviderClass': None,
            'inputStreamLabels': None,
            'inputStreams': -1,
            'inputStreamLabelProviderClass': 'com.streamsets.pipeline.api.StageDef$DefaultInputStreams',
            'connectionVerifierStage': False,
            'offsetCommitTrigger': False,
            'executionModes': ['STANDALONE', 'EDGE'],
            'resetOffset': False,
            'onlineHelpRefUrl': 'index.html?contextID=concept_czx_ktn_ht',
            'hideStage': [],
            'outputStreamsDrivenByConfig': '',
            'sendsResponse': False,
            'beta': False,
            'bisectable': False,
            'yamlUpgrader': 'upgrader/RandomDataGeneratorSource.yaml',
            'dynamicLabel': '',
            'onRecordError': True,
            'producingEvents': True,
            'preconditions': False,
        },
    ]
}

DEPLOYMENT_BUILDER_JSON = {
    'id': None,
    'organization': None,
    'name': None,
    'environment': None,
    'engineConfiguration': {
        'id': None,
        'organization': None,
        'stageLibs': None,
        'engineType': None,
        'engineVersionId': None,
        'engineVersion': None,
        'scalaBinaryVersion': None,
        'jvmConfig': {
            'id': None,
            'memoryConfigStrategy': None,
            'jvmMinMemory': None,
            'jvmMaxMemory': None,
            'jvmMinMemoryPercent': None,
            'jvmMaxMemoryPercent': None,
            'extraJvmOpts': None,
        },
        'externalResourcesUri': None,
        'advancedConfiguration': None,
        'labels': None,
        'lastStalenessTimestamp': None,
        'staleMessage': None,
        'creator': None,
        'createTime': None,
        'lastModifiedBy': None,
        'lastModifiedOn': None,
        'maxCpuLoad': None,
        'maxMemoryUsed': None,
        'maxPipelinesRunning': None,
    },
    'desiredInstances': None,
    'state': None,
    'status': None,
    'stateDisplayLabel': None,
    'statusDetail': None,
    'statusLastUpdated': None,
    'creator': None,
    'createTime': None,
    'lastModifiedBy': None,
    'lastModifiedOn': None,
    'deploymentTags': None,
    'rawDeploymentTags': None,
    'locked': None,
    'lockReference': None,
    'lockChangedAt': None,
    'lockChangedBy': None,
    'engineShutdownTimeout': None,
    'engineShutdownTimeoutTimestamp': None,
}

AZURE_ENVIRONMENT_JSON = {
    'id': None,
    'organization': None,
    'name': None,
    'type': 'AZURE',
    'userProvided': None,
    'credentialsType': None,
    'credentials': None,
    'allowSnapshotEngineVersions': None,
    'state': None,
    'status': None,
    'stateDisplayLabel': None,
    'statusDetail': None,
    'statusLastUpdated': None,
    'creator': None,
    'createTime': None,
    'lastModifiedBy': None,
    'lastModifiedOn': None,
    'environmentTags': None,
    'rawEnvironmentTags': None,
    'region': None,
    'vpcId': None,
    'subnetId': None,
    'securityGroupId': None,
    'resourceTags': None,
    'defaultResourceGroup': None,
    'defaultManagedIdentity': None,
}



# Copyright 2020 StreamSets Inc.

"""
This module contains project-wide constants.
"""

# fmt: off
import collections

from .exceptions import ConnectError, InvalidError, RunError, RunningError, StartError, StartingError, ValidationError

# fmt: on


ENGINE_AUTHENTICATION_METHOD_FORM = 'form'
ENGINE_AUTHENTICATION_METHOD_ASTER = 'aster'

SDC_EXECUTOR_TYPE = 'COLLECTOR'
SNOWFLAKE_EXECUTOR_TYPE = 'SNOWPARK'
SNOWFLAKE_ENGINE_ID = 'SYSTEM_SNOWPARK_ENGINE_ID'
TRANSFORMER_EXECUTOR_TYPE = 'TRANSFORMER'

SDC_DEPLOYMENT_TYPE = 'DC'
TRANSFORMER_DEPLOYMENT_TYPE = 'TF'

EXECUTOR_TO_DEPLOYMENT_MAP = {
    SDC_EXECUTOR_TYPE: SDC_DEPLOYMENT_TYPE,
    TRANSFORMER_EXECUTOR_TYPE: TRANSFORMER_DEPLOYMENT_TYPE,
}

ENGINELESS_ENGINE_ID = 'SYSTEM_DESIGNER'
ENGINELESS_CONNECTION_ID = 'ENGINELESS_ID'

StageConfigurationProperty = collections.namedtuple('StageConfigurationProperty', ['config_name'])
ServiceConfigurationProperty = collections.namedtuple('ServiceConfigurationProperty', ['service_name', 'config_name'])

# Default values for Deployment Engine Configurations
DEFAULT_MAX_CPU_LOAD_VALUE = 80.0
DEFAULT_MAX_MEMORY_USED_VALUE = 100.0
DEFAULT_MAX_PIPELINES_RUNNING_VALUE = 1000000


# This is a dictionary mapping stage names to attribute aliases. As an example, if a particular
# stage had a label updated in a more recent SDC release, the older attribute name should be mapped
# to the newer one.
SDC_STAGE_ATTRIBUTE_RENAME_ALIASES = {
    'com_streamsets_pipeline_stage_destination_couchbase_CouchbaseConnectorDTarget': {
        # https://git.io/fABEc
        'generate_unique_document_key': 'generate_document_key',
        'unique_document_key_field': 'document_key_field',
    },
    'com_streamsets_pipeline_stage_destination_hdfs_HdfsDTarget': {
        # https://git.io/fjCZ1, https://git.io/fjCZD
        'hadoop_fs_configuration': 'additional_configuration',
        'hadoop_fs_configuration_directory': 'configuration_files_directory',
        'hadoop_fs_uri': 'file_system_uri',
        'hdfs_user': 'impersonation_user',
        'validate_hdfs_permissions': 'validate_permissions',
    },
    'com_streamsets_pipeline_stage_destination_s3_AmazonS3DTarget': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_stage_destination_s3_S3ConnectionTargetConfig': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_stage_destination_s3_ToErrorAmazonS3DTarget': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_stage_destination_snowflake_SnowflakeDTarget': {
        # https://git.io/Jednw, https://git.io/JJ61u, https://git.io/JJ61w
        'cdc_data': 'processing_cdc_data',
        'stage_location': 'external_stage_location',
        'external_stage_name': 'snowflake_stage_name',
    },
    'com_streamsets_pipeline_stage_executor_s3_AmazonS3DExecutor': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_stage_origin_hdfs_HdfsDSource': {
        # https://git.io/fjCZ1, https://git.io/fjCZD
        'hadoop_fs_configuration': 'additional_configuration',
        'hadoop_fs_configuration_directory': 'configuration_files_directory',
        'hadoop_fs_uri': 'file_system_uri',
        'hdfs_user': 'impersonation_user',
    },
    'com_streamsets_pipeline_stage_origin_httpserver_HttpServerDPushSource': {
        # https://git.io/JvBgX
        'application_id': 'list_of_application_ids',
    },
    'com_streamsets_pipeline_stage_origin_jdbc_cdc_oracle_OracleCDCDSource': {
        # https://git.io/Jt6CY
        'use_peg_parser_in_beta': 'use_peg_parser',
    },
    'com_streamsets_pipeline_stage_origin_spooldir_SpoolDirDSource': {
        # https://git.io/vpJ8w
        'max_files_in_directory': 'max_files_soft_limit',
    },
    'com_streamsets_pipeline_stage_origin_startJob_StartJobDSource': {
        # https://git.io/JJsTN
        'delay_between_state_checks': 'status_check_interval',
        'unique_task_name': 'task_name',
        'url_of_control_hub_that_runs_the_specified_jobs': 'control_hub_url',
    },
    'com_streamsets_pipeline_stage_origin_startPipeline_StartPipelineDSource': {
        # https://git.io/JJsTN
        'delay_between_state_checks': 'status_check_interval',
        'control_hub_base_url': 'control_hub_url',
        'unique_task_name': 'task_name',
    },
    'com_streamsets_pipeline_stage_processor_kudulookup_KuduLookupDProcessor': {
        # https://git.io/vpJZF
        'ignore_missing_value': 'ignore_missing_value_in_matching_record',
    },
    'com_streamsets_pipeline_stage_processor_startJob_StartJobDProcessor': {
        # https://git.io/JJsTN
        'delay_between_state_checks': 'status_check_interval',
        'unique_task_name': 'task_name',
        'url_of_control_hub_that_runs_the_specified_jobs': 'control_hub_url',
    },
    'com_streamsets_pipeline_stage_processor_startPipeline_StartPipelineDProcessor': {
        # https://git.io/JJsTN
        'delay_between_state_checks': 'status_check_interval',
        'control_hub_base_url': 'control_hub_url',
        'unique_task_name': 'task_name',
    },
    'com_streamsets_pipeline_stage_processor_waitForJobCompletion_WaitForJobCompletionDProcessor': {
        # https://git.io/JJsTN
        'delay_between_state_checks': 'status_check_interval',
        'url_of_control_hub': 'control_hub_url',
    },
    'com_streamsets_pipeline_stage_processor_waitForPipelineCompletion_WaitForPipelineCompletionDProcessor': {
        # https://git.io/JJsTN
        'delay_between_state_checks': 'status_check_interval',
    },
    'com_streamsets_pipeline_stage_processor_controlHub_ControlHubApiDProcessor': {
        # https://git.io/JOcVh
        'control_hub_api_url': 'control_hub_url',
    },
    'com_streamsets_pipeline_stage_processor_http_HttpDProcessor': {
        'pass_record': 'pass_records',
    },
    'com_streamsets_pipeline_stage_origin_jdbc_cdc_postgres_PostgresCDCDSource': {
        # https://git.io/J1O0w
        'database_time_zone': 'db_time_zone',
    },
}

# This is a dictionary mapping stage names to attribute aliases. As an example, if a particular
# stage had a label updated in a more recent TfS release, the older attribute name should be mapped
# to the newer one.
SNOWFLAKE_STAGE_ATTRIBUTE_RENAME_ALIASES = {
    'com_streamsets_transformer_snowpark_processor_join_JoinDProcessor': {
        'input_1': 'input_1_in_left',
        'input_2': 'input_2_in_right',
        'join_using_condition': 'join_condition',
    },
}

# This dictionary maps the stage labels from old labeling to the new ones.
# When calling SDCPipelineBuilder.add_stage('Snowflake SQL Evaluator') it should return the stage with
# label Column Transformer.
SNOWFLAKE_STAGE_RENAME_ALIASES = {'Snowflake SQL Evaluator': 'Column Transformer'}

# This is a dictionary mapping stage names to attribute aliases. As an example, if a particular
# stage had a label updated in a more recent ST release, the older attribute name should be mapped
# to the newer one.
ST_STAGE_ATTRIBUTE_RENAME_ALIASES = {
    'com_streamsets_pipeline_spark_destination_redshift_RedshiftDDestination': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_spark_destination_s3_AmazonS3DDestination': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_spark_origin_jdbc_table_branded_mysql_MySQLJdbcTableDOrigin': {
        'max_number_of_partitions': 'number_of_partitions',
    },
    'com_streamsets_pipeline_spark_origin_jdbc_table_branded_oracle_OracleJdbcTableDOrigin': {
        'max_number_of_partitions': 'number_of_partitions',
    },
    'com_streamsets_pipeline_spark_origin_jdbc_table_branded_postgresql_PostgreJdbcTableDOrigin': {
        'max_number_of_partitions': 'number_of_partitions',
    },
    'com_streamsets_pipeline_spark_origin_jdbc_table_branded_sqlserver_SqlServerJdbcTableDOrigin': {
        'max_number_of_partitions': 'number_of_partitions',
    },
    'com_streamsets_pipeline_spark_origin_redshift_RedshiftDOrigin': {
        'bucket': 'bucket_and_path',
    },
    'com_streamsets_pipeline_spark_origin_s3_AmazonS3DOrigin': {
        'bucket': 'bucket_and_path',
    },
}

# For various reasons, a few configurations don't lend themselves to easy conversion
# from their labels; the overrides handle those edge cases.
STAGE_CONFIG_OVERRIDES = {
    'com_streamsets_datacollector_pipeline_executor_spark_SparkDExecutor': {
        'maximum_time_to_wait_in_ms': StageConfigurationProperty('conf.yarnConfigBean.waitTimeout'),
    },
    'com_streamsets_pipeline_stage_bigquery_origin_BigQueryDSource': {
        'use_cached_query_results': StageConfigurationProperty('conf.useQueryCache'),
    },
    'com_streamsets_pipeline_stage_destination_elasticsearch_ElasticSearchDTarget': {
        'security_username_and_password': StageConfigurationProperty('elasticSearchConfig.securityConfig.securityUser'),
    },
    'com_streamsets_pipeline_stage_destination_elasticsearch_ToErrorElasticSearchDTarget': {
        'security_username_and_password': StageConfigurationProperty('elasticSearchConfig.securityConfig.securityUser'),
    },
    'com_streamsets_pipeline_stage_destination_snowflake_SnowflakeDTarget': {
        'processing_cdc_data': StageConfigurationProperty('config.data.cdcData'),
    },
    'com_streamsets_pipeline_stage_devtest_rawdata_RawDataDSource': {
        'data_format': [
            StageConfigurationProperty('dataFormat'),
            ServiceConfigurationProperty(
                'com.streamsets.pipeline.api.service.' 'dataformats.DataFormatParserService', 'dataFormat'
            ),
        ],
        'datagram_data_format': [
            StageConfigurationProperty('dataFormatConfig.datagramMode'),
            ServiceConfigurationProperty(
                'com.streamsets.pipeline.api.service.' 'dataformats.DataFormatParserService',
                'dataFormatConfig.datagramMode',
            ),
        ],
    },
    'com_streamsets_pipeline_stage_origin_coapserver_CoapServerDPushSource': {
        'data_format': StageConfigurationProperty('dataFormat'),
        'datagram_data_format': StageConfigurationProperty('dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_elasticsearch_ElasticsearchDSource': {
        'security_username_and_password': StageConfigurationProperty('elasticSearchConfig.securityConfig.securityUser'),
    },
    'com_streamsets_pipeline_stage_origin_hdfs_cluster_ClusterHdfsDSource': {
        'data_format': StageConfigurationProperty('clusterHDFSConfigBean.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('clusterHDFSConfigBean.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_http_HttpClientDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
        'initial_page_or_offset': StageConfigurationProperty('conf.pagination.startAt'),
    },
    'com_streamsets_pipeline_stage_origin_httpserver_HttpServerDPushSource': {
        'data_format': StageConfigurationProperty('dataFormat'),
        'datagram_data_format': StageConfigurationProperty('dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_jms_JmsDSource': {
        'data_format': [
            StageConfigurationProperty('dataFormat'),
            ServiceConfigurationProperty(
                'com.streamsets.pipeline.api.service.' 'dataformats.DataFormatParserService', 'dataFormat'
            ),
        ],
        'datagram_data_format': [
            StageConfigurationProperty('dataFormatConfig.datagramMode'),
            ServiceConfigurationProperty(
                'com.streamsets.pipeline.api.service.' 'dataformats.DataFormatParserService',
                'dataFormatConfig.datagramMode',
            ),
        ],
    },
    'com_streamsets_pipeline_stage_origin_kafka_KafkaDSource': {
        'data_format': StageConfigurationProperty('kafkaConfigBean.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('kafkaConfigBean.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_kinesis_KinesisDSource': {
        'data_format': StageConfigurationProperty('kinesisConfig.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('kinesisConfig.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_logtail_FileTailDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_maprfs_ClusterMapRFSDSource': {
        'data_format': StageConfigurationProperty('clusterHDFSConfigBean.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('clusterHDFSConfigBean.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_maprstreams_MapRStreamsDSource': {
        'data_format': StageConfigurationProperty('maprstreamsSourceConfigBean.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('maprstreamsSourceConfigBean.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_mqtt_MqttClientDSource': {
        'data_format': StageConfigurationProperty('subscriberConf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('subscriberConf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_rabbitmq_RabbitDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_redis_RedisDSource': {
        'data_format': StageConfigurationProperty('redisOriginConfigBean.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('redisOriginConfigBean.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_remote_RemoteDownloadDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_s3_AmazonS3DSource': {
        'data_format': [
            StageConfigurationProperty('s3ConfigBean.dataFormat'),
            ServiceConfigurationProperty(
                'com.streamsets.pipeline.api.service.' 'dataformats.DataFormatParserService', 'dataFormat'
            ),
        ],
    },
    'com_streamsets_pipeline_stage_origin_spooldir_SpoolDirDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_tcp_TCPServerDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_origin_websocketserver_WebSocketServerDPushSource': {
        'data_format': StageConfigurationProperty('dataFormat'),
        'datagram_data_format': StageConfigurationProperty('dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_processor_http_HttpDProcessor': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
    'com_streamsets_pipeline_stage_pubsub_origin_PubSubDSource': {
        'data_format': StageConfigurationProperty('conf.dataFormat'),
        'datagram_data_format': StageConfigurationProperty('conf.dataFormatConfig.datagramMode'),
    },
}

STATUS_ERRORS = {
    'CONNECT_ERROR': ConnectError,
    'RUN_ERROR': RunError,
    'RUNNING_ERROR': RunningError,
    'START_ERROR': StartError,
    'STARTING_ERROR': StartingError,
    'INVALID': InvalidError,
    'VALIDATION_ERROR': ValidationError,
}


"""
 The following dictionary keeps backward compatibility for renamed pipeline configurations in Transformer engines.
 The way this dictionary should be formatted is as following:

    {
        'version1': {
            'old name 1' {
                'name': 'new name 1',
                'values': {
                    'old value 1': 'new value 1',
                    'old value 2': 'new value 2'
                }
            },
            'old value 2': {
                'name': 'new name 2',
                'values': {
                    'old value 3': 'new value 3',
                    'old value 4': 'new value 4'
                }
            }
        },
        'more versions': {
            ...
        }
    }

 The versions should be sorted in ascending order.
"""

ST_PIPELINE_BW_COMPATIBILITY = {
    '5.7.0': {
        'transformerEmrConnection.provisionNewCluster': {
            'name': 'transformerEmrConnection.emrClusterOption',
            'values': {True: 'PROVISION_CLUSTER', False: 'EXISTING_CLUSTER'},
        },
        'transformerEmrConnection.createNewApplication': {
            'name': 'transformerEmrConnection.emrClusterOption',
            'values': {True: 'PROVISION_CLUSTER', False: 'EXISTING_CLUSTER'},
        },
    },
    '6.0.0': {
        # Transformer
        'transformerEmrConnection.masterInstanceType': {'name': 'transformerEmrConnection.primaryInstanceType'},
        'transformerEmrConnection.masterInstanceTypeCustom': {
            'name': 'transformerEmrConnection.primaryInstanceTypeCustom'
        },
        'transformerEmrConnection.masterSecurityGroup': {'name': 'transformerEmrConnection.primarySecurityGroup'},
        'transformerEmrConnection.slaveInstanceType': {'name': 'transformerEmrConnection.secondaryInstanceType'},
        'transformerEmrConnection.slaveInstanceTypeCustom': {
            'name': 'transformerEmrConnection.secondaryInstanceTypeCustom'
        },
        'transformerEmrConnection.slaveSecurityGroup': {'name': 'transformerEmrConnection.secondarySecurityGroup'},
        # Data Collector
        'sdcEmrConnection.masterInstanceType': {'name': 'sdcEmrConnection.primaryInstanceType'},
        'sdcEmrConnection.masterInstanceTypeCustom': {'name': 'sdcEmrConnection.primaryInstanceTypeCustom'},
        'sdcEmrConnection.masterSecurityGroup': {'name': 'sdcEmrConnection.primarySecurityGroup'},
        'sdcEmrConnection.slaveInstanceType': {'name': 'sdcEmrConnection.secondaryInstanceType'},
        'sdcEmrConnection.slaveInstanceTypeCustom': {'name': 'sdcEmrConnection.secondaryInstanceTypeCustom'},
        'sdcEmrConnection.slaveSecurityGroup': {'name': 'sdcEmrConnection.secondarySecurityGroup'},
    },
}


"""
 The following dictionary keeps backward compatibility for renamed pipeline configurations in Transformer engines.
 This dictionary should be formatted is the same as the pipeline compatibility map,
 with the connection type as a top level condition.
 The versions should be sorted inside each connection type in ascending order.
"""


SDC_CONNECTIONS_BW_COMPATIBILITY = {
    'STREAMSETS_AWS_EMR_CLUSTER': {
        '5.10.0': {
            'provisionNewCluster': {
                'name': 'emrClusterOption',
                'values': {True: 'PROVISION_CLUSTER', False: 'EXISTING_CLUSTER'},
            },
            'createNewApplication': {
                'name': 'emrClusterOption',
                'values': {True: 'PROVISION_CLUSTER', False: 'EXISTING_CLUSTER'},
            },
        }
    },
}



# Copyright 2023 StreamSets Inc.

# fmt: off
import pytest

from streamsets.sdk.sch_models import ScheduledTaskStatus
from streamsets.sdk.utils import get_random_string

# fmt: on


@pytest.fixture(scope="module")
def sample_job(sch, sch_authoring_sdc_id, sch_executor_sdc_label):
    """A trivial job

    dev_data_generator >> trash
    """
    pipeline_builder = sch.get_pipeline_builder(engine_type='data_collector', engine_id=sch_authoring_sdc_id)
    dev_data_generator = pipeline_builder.add_stage('Dev Data Generator')
    trash = pipeline_builder.add_stage('Trash')
    dev_data_generator >> trash
    pipeline = pipeline_builder.build('simple_pipeline_{}'.format(get_random_string()))
    sch.publish_pipeline(pipeline)

    job_builder = sch.get_job_builder()
    job = job_builder.build(
        'Job Template SDK Tests {}'.format(get_random_string()),
        pipeline=pipeline,
        runtime_parameters={'x': 'y', 'a': 'b'},
    )
    sch.add_job(job)

    try:
        yield job
    finally:
        sch.delete_job(job)
        sch.api_client.delete_pipeline(pipeline.pipeline_id)


@pytest.fixture
def sample_running_scheduled_task(sch, sample_job):
    scheduled_task_builder = sch.get_scheduled_task_builder()
    scheduled_task = scheduled_task_builder.build(sample_job)
    sch.add_scheduled_task(scheduled_task)

    try:
        yield scheduled_task
    finally:
        if ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.DELETED:
            return
        if ScheduledTaskStatus(scheduled_task.status) != ScheduledTaskStatus.KILLED:
            sch.kill_scheduled_tasks(scheduled_task)
        sch.delete_scheduled_tasks(scheduled_task)


@pytest.fixture
def sample_paused_scheduled_task(sch, sample_job):
    scheduled_task_builder = sch.get_scheduled_task_builder()
    scheduled_task = scheduled_task_builder.build(sample_job)
    sch.add_scheduled_task(scheduled_task)
    sch.pause_scheduled_tasks(scheduled_task)

    try:
        yield scheduled_task
    finally:
        if ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.DELETED:
            return
        if ScheduledTaskStatus(scheduled_task.status) != ScheduledTaskStatus.KILLED:
            sch.kill_scheduled_tasks(scheduled_task)
        sch.delete_scheduled_tasks(scheduled_task)


@pytest.fixture
def sample_killed_scheduled_task(sch, sample_job):
    scheduled_task_builder = sch.get_scheduled_task_builder()
    scheduled_task = scheduled_task_builder.build(sample_job)
    sch.add_scheduled_task(scheduled_task)
    sch.kill_scheduled_tasks(scheduled_task)

    try:
        yield scheduled_task
    finally:
        if ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.DELETED:
            return
        if ScheduledTaskStatus(scheduled_task.status) != ScheduledTaskStatus.KILLED:
            sch.kill_scheduled_tasks(scheduled_task)
        sch.delete_scheduled_tasks(scheduled_task)


@pytest.fixture
def sample_deleted_scheduled_task(sch, sample_job):
    scheduled_task_builder = sch.get_scheduled_task_builder()
    scheduled_task = scheduled_task_builder.build(sample_job)
    sch.add_scheduled_task(scheduled_task)
    sch.kill_scheduled_tasks(scheduled_task)
    sch.delete_scheduled_tasks(scheduled_task)

    try:
        yield scheduled_task
    finally:
        if ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.DELETED:
            return
        if ScheduledTaskStatus(scheduled_task.status) != ScheduledTaskStatus.KILLED:
            sch.kill_scheduled_tasks(scheduled_task)
        sch.delete_scheduled_tasks(scheduled_task)


def test_scheduled_task_in_sch(sch, sample_running_scheduled_task):
    assert sample_running_scheduled_task in sch.scheduled_tasks


def test_scheduled_task_removed_from_sch(sch, sample_deleted_scheduled_task):
    assert sample_deleted_scheduled_task not in sch.scheduled_tasks


def test_overriden_contains_function(sch, sample_running_scheduled_task):
    task_id = sample_running_scheduled_task.id
    task_name = sample_running_scheduled_task.name
    # should be present
    assert sch.scheduled_tasks.contains(id=task_id)
    assert sch.scheduled_tasks.contains(name=task_name)
    # should not be present
    assert not sch.scheduled_tasks.contains(id=get_random_string())
    assert not sch.scheduled_tasks.contains(name=get_random_string())


@pytest.mark.parametrize(
    "scheduled_task",
    [
        'sample_running_scheduled_task',
        'sample_paused_scheduled_task',
        'sample_killed_scheduled_task',
        'sample_deleted_scheduled_task',
    ],
)
def test_resume_action(sch, scheduled_task, request):
    scheduled_task = request.getfixturevalue(scheduled_task)
    initial_status = ScheduledTaskStatus(scheduled_task.status)
    sch.resume_scheduled_tasks(scheduled_task)
    if initial_status == ScheduledTaskStatus.PAUSED:
        assert ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.RUNNING
    else:
        assert ScheduledTaskStatus(scheduled_task.status) == initial_status


@pytest.mark.parametrize(
    "scheduled_task",
    [
        'sample_running_scheduled_task',
        'sample_paused_scheduled_task',
        'sample_killed_scheduled_task',
        'sample_deleted_scheduled_task',
    ],
)
def test_pause_action(sch, scheduled_task, request):
    scheduled_task = request.getfixturevalue(scheduled_task)
    initial_status = ScheduledTaskStatus(scheduled_task.status)
    sch.pause_scheduled_tasks(scheduled_task)

    if initial_status == ScheduledTaskStatus.RUNNING:
        assert ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.PAUSED
    else:
        assert ScheduledTaskStatus(scheduled_task.status) == initial_status


@pytest.mark.parametrize(
    "scheduled_task",
    [
        'sample_running_scheduled_task',
        'sample_paused_scheduled_task',
        'sample_killed_scheduled_task',
        'sample_deleted_scheduled_task',
    ],
)
def test_kill_action(sch, scheduled_task, request):
    scheduled_task = request.getfixturevalue(scheduled_task)
    initial_status = ScheduledTaskStatus(scheduled_task.status)
    sch.kill_scheduled_tasks(scheduled_task)
    if initial_status in [ScheduledTaskStatus.RUNNING, ScheduledTaskStatus.PAUSED]:
        assert ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.KILLED
    else:
        assert ScheduledTaskStatus(scheduled_task.status) == initial_status


@pytest.mark.parametrize(
    "scheduled_task",
    [
        'sample_running_scheduled_task',
        'sample_paused_scheduled_task',
        'sample_killed_scheduled_task',
        'sample_deleted_scheduled_task',
    ],
)
def test_delete_action(sch, scheduled_task, request):
    scheduled_task = request.getfixturevalue(scheduled_task)
    initial_status = ScheduledTaskStatus(scheduled_task.status)
    sch.delete_scheduled_tasks(scheduled_task)
    if initial_status == ScheduledTaskStatus.KILLED:
        assert ScheduledTaskStatus(scheduled_task.status) == ScheduledTaskStatus.DELETED
    else:
        assert ScheduledTaskStatus(scheduled_task.status) == initial_status


@pytest.mark.parametrize(
    "function_name, result_status, first_task, second_task",
    [
        (
            'resume_scheduled_tasks',
            ScheduledTaskStatus.RUNNING,
            'sample_paused_scheduled_task',
            'sample_paused_scheduled_task',
        ),
        (
            'pause_scheduled_tasks',
            ScheduledTaskStatus.PAUSED,
            'sample_running_scheduled_task',
            'sample_running_scheduled_task',
        ),
        (
            'kill_scheduled_tasks',
            ScheduledTaskStatus.KILLED,
            'sample_running_scheduled_task',
            'sample_paused_scheduled_task',
        ),
        (
            'delete_scheduled_tasks',
            ScheduledTaskStatus.DELETED,
            'sample_killed_scheduled_task',
            'sample_killed_scheduled_task',
        ),
    ],
)
def test_action_on_multiple_scheduled_tasks(sch, function_name, result_status, first_task, second_task, request):
    first_task = request.getfixturevalue(first_task)
    second_task = request.getfixturevalue(second_task)
    function_to_test = getattr(sch, function_name)
    function_to_test(first_task, second_task)
    assert ScheduledTaskStatus(first_task.status) == result_status
    assert ScheduledTaskStatus(second_task.status) == result_status


@pytest.mark.parametrize(
    "function_name, first_task, first_task_result_status, second_task, second_task_result_status",
    [
        (
            'resume_scheduled_tasks',
            'sample_killed_scheduled_task',
            ScheduledTaskStatus.KILLED,
            'sample_paused_scheduled_task',
            ScheduledTaskStatus.RUNNING,
        ),
        (
            'pause_scheduled_tasks',
            'sample_killed_scheduled_task',
            ScheduledTaskStatus.KILLED,
            'sample_running_scheduled_task',
            ScheduledTaskStatus.PAUSED,
        ),
        (
            'kill_scheduled_tasks',
            'sample_deleted_scheduled_task',
            ScheduledTaskStatus.DELETED,
            'sample_running_scheduled_task',
            ScheduledTaskStatus.KILLED,
        ),
        (
            'delete_scheduled_tasks',
            'sample_paused_scheduled_task',
            ScheduledTaskStatus.PAUSED,
            'sample_killed_scheduled_task',
            ScheduledTaskStatus.DELETED,
        ),
    ],
)
def test_skip_tasks_not_permissible_state(
    sch, function_name, first_task, first_task_result_status, second_task, second_task_result_status, request
):
    # only one of the tasks is in the correct state, the first task should be skipped by the function
    first_task = request.getfixturevalue(first_task)
    second_task = request.getfixturevalue(second_task)
    function_to_test = getattr(sch, function_name)
    function_to_test(first_task, second_task)
    assert ScheduledTaskStatus(first_task.status) == first_task_result_status
    assert ScheduledTaskStatus(second_task.status) == second_task_result_status


@pytest.mark.parametrize(
    "function_name",
    ['resume_scheduled_tasks', 'pause_scheduled_tasks', 'kill_scheduled_tasks', 'delete_scheduled_tasks'],
)
def test_call_function_without_parameters(sch, function_name):
    function_to_test = getattr(sch, function_name)
    with pytest.raises(ValueError):
        function_to_test()


@pytest.mark.parametrize(
    "function_name",
    ['resume_scheduled_tasks', 'pause_scheduled_tasks', 'kill_scheduled_tasks', 'delete_scheduled_tasks'],
)
@pytest.mark.parametrize("parameters", [(1, 2, 3), ('hello world', "this shouldn't work", get_random_string()), 'wow'])
def test_passing_random_variables_parameters(sch, function_name, parameters):
    function_to_test = getattr(sch, function_name)
    with pytest.raises(ValueError):
        function_to_test(*parameters)


@pytest.mark.parametrize("random_object", ['hello world', 4, ScheduledTaskStatus.DELETED])
def test_passing_random_variables_in_action(sch, sample_running_scheduled_task, random_object):
    with pytest.raises(ValueError):
        sch._validate_and_execute_scheduled_tasks_action(random_object, sample_running_scheduled_task)


def test_update_scheduled_task(sch, sample_running_scheduled_task):
    new_name = 'Switched Name'
    task_id = sample_running_scheduled_task.id
    task = sch.scheduled_tasks.get(id=task_id)
    task.name = new_name
    sch.update_scheduled_task(task)
    task = sch.scheduled_tasks.get(id=task_id)
    assert task.name == new_name


def test_filter_by(sch, sample_running_scheduled_task):
    results = sch.scheduled_tasks.get_all(filter_by=sample_running_scheduled_task.name)
    assert len(results) == 1

    results = sch.scheduled_tasks.get_all(filter_by='NAME THAT CLEARLY DOESNT EXIST')
    assert len(results) == 0



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2024

import inspect
import sys

FUNCTION_CALLS_HEADERS_KEY = 'X-SS-Sdk-Functions'
INTERACTIVE_MODE_HEADERS_KEY = 'X-SS-Sdk-Interactive-mode'
SDK_FILE_PATH = 'python/streamsets/sdk'
MAX_THRESHOLD = 500


class AnalyticsHeaders:
    """Class to handle storing functions called during the Users session.

    This class stores the data in a dict of the form:
        {'X-SS-Sdk-Functions': <str of function name>,
         'X-SS-Sdk-Interactive-mode': <bool of whether in interactive mode>}
    """

    # maintain an _instance to enable this class to function as a Singleton
    _instance = None

    def __init__(self):
        self.interactive_mode = hasattr(sys, 'ps2')

        self.headers = None
        self.initalize_empty_headers()

    def initalize_empty_headers(self):
        self.headers = {FUNCTION_CALLS_HEADERS_KEY: None, INTERACTIVE_MODE_HEADERS_KEY: str(self.interactive_mode)}

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = AnalyticsHeaders()
        return cls._instance

    @classmethod
    def is_interactive_mode(cls):
        return cls.get_instance().interactive_mode

    @classmethod
    def add_attribute_to_headers(cls, class_obj, name):
        attribute_name = "{}.{}".format(class_obj.__name__, name)
        instance = cls.get_instance()
        instance.headers[FUNCTION_CALLS_HEADERS_KEY] = attribute_name

    @classmethod
    def reset_headers(cls):
        instance = cls.get_instance()
        instance.initalize_empty_headers()


def analytics_class_decorator(cls):
    """Class decorator that adds function user calls during their session to AnalyticsHeaders.

    This decorator overrides:
        1. __getattribute__: This ensures that every method, property and attribute that gets called goes through
        new_getattribute and can accurately be tracked.

        2. __setattr__: This ensures that setting any attribute/property goes through new_setattr and can accurately
         be tracked.

    Both methods go through the following logic:
    1. If the call is made from the user, track via AnalyticsHeaders.add_attribute_to_headers
    2. Call the original __getattribute__ or __setattr__
    """
    original_setattr = cls.__setattr__
    original_getattribute = cls.__getattribute__

    def new_setattr(self, name, value):
        """Overloads original __setattr__ to add metrics data."""
        # Make sure that func was called from the user script and not from an internal method
        caller_code = inspect.currentframe().f_back.f_code
        analytics_instance = AnalyticsHeaders.get_instance()
        if SDK_FILE_PATH not in caller_code.co_filename:
            analytics_instance.reset_headers()
            AnalyticsHeaders.add_attribute_to_headers(cls, name)

        return original_setattr(self, name, value)

    def new_getattribute(self, name):
        """Overloads original __getattribute__ to add metrics data."""
        # Make sure that func was called from the user script and not from an internal method
        caller_code = inspect.currentframe().f_back.f_code
        analytics_instance = AnalyticsHeaders.get_instance()
        if SDK_FILE_PATH not in caller_code.co_filename:
            analytics_instance.reset_headers()
            AnalyticsHeaders.add_attribute_to_headers(cls, name)

        return original_getattribute(self, name)

    cls.__setattr__ = new_setattr
    cls.__getattribute__ = new_getattribute

    return cls
